# SPDX-License-Identifier: MIT

import struct
from io import BytesIO

from enum import IntEnum

from m1n1.constructutils import Ver
from m1n1.proxyutils import RegMonitor
from m1n1.utils import *
from m1n1.trace.dart import DARTTracer
from m1n1.trace.asc import ASCTracer, EP, EPState, msg, msg_log, DIR
from m1n1.fw.afk.rbep import *
from m1n1.fw.afk.epic import *

Ver.set_version(hv.u)

class DCPDevType(IntEnum):
    DCP = 0
    DART_DCP = 1
    DART_DISP = 2

def get_alias(node):
    if "aliases" in u.adt and hasattr(u.adt["aliases"], node):
        return u.adt["aliases"].getprop(node).rsplit('/', 1)[1]
    return node

def get_dcp_device(dev_type, dcp):
    if dev_type is DCPDevType.DCP:
        return "/arm-io/" + get_alias(dcp)
    elif dev_type is DCPDevType.DART_DCP:
        return "/arm-io/dart-" + get_alias(dcp)
    elif dev_type is DCPDevType.DART_DISP:
        return "/arm-io/dart-" + dcp.replace("dcp", "disp")
    else:
        raise KeyError(dev_type)

dcp_name = "dcp0"
#dcp_name = "dcpext0"

trace_device(get_dcp_device(DCPDevType.DCP, dcp_name), True, ranges=[1])

DARTTracer = DARTTracer._reloadcls()
ASCTracer = ASCTracer._reloadcls()

iomon = RegMonitor(hv.u, ascii=True)

class AFKRingBufSniffer(AFKRingBuf):
    def __init__(self, ep, state, base, size):
        super().__init__(ep, base, size)
        self.state = state
        self.rptr = getattr(state, "rptr", 0)

    def update_rptr(self, rptr):
        self.state.rptr = rptr

    def update_wptr(self):
        raise NotImplementedError()

    def get_wptr(self):
        return struct.unpack("<I", self.read_buf(2 * self.BLOCK_STEP, 4))[0]

    def read_buf(self, off, size):
        return self.ep.dart.ioread(self.ep.stream, self.base + off, size)

class AFKEp(EP):
    BASE_MESSAGE = AFKEPMessage

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.txbuf = None
        self.rxbuf = None
        self.state.txbuf = EPState()
        self.state.rxbuf = EPState()
        self.state.shmem_iova = None
        self.state.txbuf_info = None
        self.state.rxbuf_info = None
        self.state.verbose = 1

    def start(self):
        #self.add_mon()
        self.create_bufs()

    def create_bufs(self):
        if not self.state.shmem_iova:
            return
        if not self.txbuf and self.state.txbuf_info:
            off, size = self.state.txbuf_info
            self.txbuf = AFKRingBufSniffer(self, self.state.txbuf,
                                           self.state.shmem_iova + off, size)
        if not self.rxbuf and self.state.rxbuf_info:
            off, size = self.state.rxbuf_info
            self.rxbuf = AFKRingBufSniffer(self, self.state.rxbuf,
                                           self.state.shmem_iova + off, size)

    def add_mon(self):
        if self.state.shmem_iova:
            iomon.add(self.state.shmem_iova, 32768,
                      name=f"{self.name}.shmem@{self.state.shmem_iova:08x}", offset=0)

    Init =          msg_log(0x80, DIR.TX)
    Init_Ack =      msg_log(0xa0, DIR.RX)

    GetBuf =        msg_log(0x89, DIR.RX)

    Shutdown =      msg_log(0xc0, DIR.TX)
    Shutdown_Ack =  msg_log(0xc1, DIR.RX)

    @msg(0xa1, DIR.TX, AFKEP_GetBuf_Ack)
    def GetBuf_Ack(self, msg):
        self.state.shmem_iova = msg.DVA
        self.txbuf = None
        self.rxbuf = None
        self.state.txbuf = EPState()
        self.state.rxbuf = EPState()
        self.state.txbuf_info = None
        self.state.rxbuf_info = None
        #self.add_mon()

    @msg(0xa2, DIR.TX, AFKEP_Send)
    def Send(self, msg):
        self.log(f">TX msg:{msg}")
        for data in self.txbuf.read():
            if self.state.verbose >= 3:
                self.log(f">TX rptr={self.txbuf.state.rptr:#x}")
                chexdump(data, print_fn=self.log)
            self.handle_ipc(data, dir=">")
        return True

    Hello =         msg_log(0xa3, DIR.TX)

    @msg(0x85, DIR.RX, AFKEPMessage)
    def Recv(self, msg):
        self.log(f"<RX msg:{msg}")
        for data in self.rxbuf.read():
            if self.state.verbose >= 3:
                self.log(f"<RX rptr={self.rxbuf.state.rptr:#x}")
                chexdump(data, print_fn=self.log)
            self.handle_ipc(data, dir="<")
        return True

    def handle_ipc(self, data, dir=None):
        pass

    @msg(0x8a, DIR.RX, AFKEP_InitRB)
    def InitTX(self, msg):
        off = msg.OFFSET * AFKRingBuf.BLOCK_STEP
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP
        self.state.txbuf_info = (off, size)
        self.create_bufs()

    @msg(0x8b, DIR.RX, AFKEP_InitRB)
    def InitRX(self, msg):
        off = msg.OFFSET * AFKRingBuf.BLOCK_STEP
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP
        self.state.rxbuf_info = (off, size)
        self.create_bufs()

class SilentEp(AFKEp):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state.verbose = 0

    def log(self, msg):
        pass

def epic_service_cmd(group, cmd):
    def f(x):
        x.is_cmd = True
        x.group = group
        x.cmd = cmd
        return x
    return f

def epic_service_reply(group, cmd):
    def f(x):
        x.is_reply = True
        x.group = group
        x.cmd = cmd
        return x
    return f

def epic_service_notify(group, cmd):
    def f(x):
        x.is_notify = True
        x.group = group
        x.cmd = cmd
        return x
    return f

def epic_service_notify_ack(group, cmd):
    def f(x):
        x.is_notify_ack = True
        x.group = group
        x.cmd = cmd
        return x
    return f

class EPICServiceTracer(Reloadable):
    def __init__(self, tracer, ep, key):
        self.tracer = tracer
        self.ep = ep
        self.key = key

        self.cmdmap = {}
        self.replymap = {}
        self.notifymap = {}
        self.notifyackmap = {}
        for name in dir(self):
            i = getattr(self, name)
            if not callable(i):
                continue
            if getattr(i, "is_cmd", False):
                self.cmdmap[i.group, i.cmd] = getattr(self, name)
            if getattr(i, "is_reply", False):
                self.replymap[i.group, i.cmd] = getattr(self, name)
            if getattr(i, "is_notify", False):
                self.notifymap[i.group, i.cmd] = getattr(self, name)
            if getattr(i, "is_notify_ack", False):
                self.notifyackmap[i.group, i.cmd] = getattr(self, name)

    def log(self, msg):
        self.ep.log(f"[{self.key}] {msg}")

    def init(self, props):
        pass

    def handle_cmd(self, sgroup, scmd, sdata):
        cmdfn = self.cmdmap.get((sgroup, scmd), None)
        if cmdfn:
            cmdfn(sdata)
        else:
            self.log(f"> unknown group {sgroup}; command {scmd}")
            if sdata:
                chexdump(sdata, print_fn=self.log)

    def handle_reply(self, sgroup, scmd, sdata):
        replyfn = self.replymap.get((sgroup, scmd), None)
        if replyfn:
            replyfn(sdata)
        else:
            self.log(f"< unknown group {sgroup}; command {scmd}")
            if sdata:
                chexdump(sdata, print_fn=self.log)

    def handle_notify(self, sgroup, scmd, sdata):
        notifyfn = self.notifymap.get((sgroup, scmd), None)
        if notifyfn:
            notifyfn(sdata)
        else:
            self.log(f"< unknown group {sgroup}; command {scmd}")
            if sdata:
                chexdump(sdata, print_fn=self.log)

    def handle_notify_ack(self, sgroup, scmd, sdata):
        notifyackfn = self.notifyackmap.get((sgroup, scmd), None)
        if notifyackfn:
            notifyackfn(sdata)
        else:
            self.log(f"< unknown group {sgroup}; command {scmd}")
            if sdata:
                chexdump(sdata, print_fn=self.log)

    @epic_service_cmd(4, 4)
    def getLocation(self, data):
        self.log("> getLocation")
    @epic_service_reply(4, 4)
    def getLocation_reply(self, data):
        self.log("< getLocation")

    @epic_service_cmd(4, 5)
    def getUnit(self, data):
        self.log("> getUnit")
    @epic_service_reply(4, 5)
    def getUnit_reply(self, data):
        self.log("< getUnit")

    @epic_service_cmd(4, 6)
    def open(self, data):
        self.log("> open")
    @epic_service_reply(4, 6)
    def open_reply(self, data):
        self.log("< open")

    @epic_service_cmd(4, 7)
    def close(self, data):
        self.log("> close")
    @epic_service_reply(4, 7)
    def close_reply(self, data):
        self.log("< close")

class EPICEp(AFKEp):
    SERVICES = []

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)

        self.serv_map = {}
        self.chan_map = {}
        self.serv_names = {}
        for i in self.SERVICES:
            self.serv_names[i.NAME] = i

    def handle_ipc(self, data, dir=None):
        fd = BytesIO(data)
        hdr = EPICHeader.parse_stream(fd)
        sub = EPICSubHeader.parse_stream(fd)

        self.log(f"{dir}Ch {hdr.channel} Type {hdr.type} Ver {hdr.version} Tag {hdr.seq}")
        self.log(f"  Len {sub.length} Ver {sub.version} Cat {sub.category} Type {int(sub.type):#x}/{sub.type} Seq {sub.seq}")
        # chexdump(data, print_fn=self.log)

        if sub.category == EPICCategory.REPORT:
            self.handle_report(hdr, sub, fd)
        elif sub.category == EPICCategory.NOTIFY:
            self.handle_notify(hdr, sub, fd)
        elif sub.category == EPICCategory.REPLY:
            if hdr.type == EPICType.NOTIFY_ACK:
                self.handle_notify_ack(hdr, sub, fd)
            else:
                self.handle_reply(hdr, sub, fd)
        elif sub.category == EPICCategory.COMMAND:
            self.handle_cmd(hdr, sub, fd)

    def handle_report_init(self, hdr, sub, fd):
            init = EPICAnnounce.parse_stream(fd)
            self.log(f"Init: {init.name}")
            self.log(f"  Props: {init.props}")

            if not init.props:
                init.props = {}

            name = init.props.get("EPICName", init.name)
            key = name + str(init.props.get("EPICUnit", ""))
            self.log(f"New service: {key} on channel {hdr.channel}")

            srv_cls = self.serv_names.get(name, EPICServiceTracer)
            srv = srv_cls(self.tracer, self, key)
            srv.init(init.props)
            srv.chan = hdr.channel
            self.chan_map[hdr.channel] = srv
            self.serv_map[key] = srv

    def handle_report(self, hdr, sub, fd):
        if sub.type == 0x30:
            self.handle_report_init(hdr, sub, fd)
        else:
            self.log(f"Report {sub.type:#x}")
            chexdump(fd.read(), print_fn=self.log)

    def handle_notify(self, hdr, sub, fd):
        self.log(f"Notify:")
        data = fd.read()
        rgroup, rcmd, rlen, rmagic = struct.unpack("<2xHIII", data[4:20])
        if rmagic != 0x69706378:
            self.log("Warning: Invalid EPICStandardService notify magic")
        srv = self.chan_map.get(hdr.channel, None)
        if srv:
            srv.handle_notify(rgroup, rcmd, data[4:] if rlen else None)
        else:
            self.log(f"[???] < group {rgroup} command {rcmd}")
            chexdump(data, print_fn=lambda msg: self.log(f"[???] {msg}"))

    def handle_notify_ack(self, hdr, sub, fd):
        self.log(f"NotifyAck:")
        data = fd.read()
        rgroup, rcmd, rlen, rmagic = struct.unpack("<2xHIII", data[4:20])
        if rmagic != 0x69706378:
            self.log("Warning: Invalid EPICStandardService notify magic")
        srv = self.chan_map.get(hdr.channel, None)
        if srv:
            srv.handle_notify_ack(rgroup, rcmd, data[4:] if rlen else None)
        else:
            self.log(f"[???] > group {rgroup} command {rcmd}")
            chexdump(data, print_fn=lambda msg: self.log(f"[???] {msg}"))

    def handle_reply(self, hdr, sub, fd):
        if sub.inline_len:
            payload = fd.read()
            self.log("Inline payload:")
            chexdump(payload, print_fn=self.log)
        else:
            off = fd.tell()
            data = fd.read()
            if len(data) == 4:
                retcode = struct.unpack("<I", data)[0]
                self.log(f"EPIC: retcode only reply retcode={retcode:#x}")
                return
            fd.seek(off)
            cmd = EPICCmd.parse_stream(fd)
            if not cmd.rxbuf:
                self.log(f"Response {sub.type:#x}: {cmd.retcode:#x}")
                return

            data = self.dart.ioread(self.stream, cmd.rxbuf, cmd.rxlen)
            if len(data) < 16:
                self.log(f"EPIC: short reply, len={len(data)}")
                chexdump(data, print_fn=self.log)
                return
            rgroup, rcmd, rlen, rmagic = struct.unpack("<2xHIII", data[:16])
            if rmagic != 0x69706378:
                self.log("Warning: Invalid EPICStandardService response magic")

            srv = self.chan_map.get(hdr.channel, None)
            if srv:
                if rgroup == 8 and rcmd == 8 and (rlen < 32 or (len(data) - 64) < 32):
                    self.log(f"hotPlugDetectChangeOccurred: rlen:{rlen} len(data):{len(data)} rxlen:{cmd.rxlen}\ncmd:{cmd}")
                    chexdump(data, print_fn=self.log)
                srv.handle_reply(rgroup, rcmd, data[64:64+rlen] if rlen else None)
            else:
                self.log(f"[???] < group {rgroup} command {rcmd}")
                chexdump(data[64:64+rlen], print_fn=lambda msg: self.log(f"[???] {msg}"))

    def handle_cmd(self, hdr, sub, fd):
        cmd = EPICCmd.parse_stream(fd)
        payload = fd.read()

        if sub.type == 0xc0 and cmd.txbuf:
            data = self.dart.ioread(self.stream, cmd.txbuf, cmd.txlen)
            if len(data) < 64:
                self.log(f"EPIC: short cmd, len={len(data)}")
                chexdump(data, print_fn=self.log)
                return
            sgroup, scmd, slen, sfooter  = struct.unpack("<2xHIII48x", data[:64])
            sdata = data[64:64+slen] if slen else None

            srv = self.chan_map.get(hdr.channel, None)
            if srv:
                srv.handle_cmd(sgroup, scmd, sdata)
            else:
                self.log(f"[???] > group {sgroup} command {scmd}")
                chexdump(data[64:64+slen], print_fn=lambda msg: self.log(f"[???] {msg}"))
        else:
            self.log(f"Command {sub.type:#x}: {cmd.retcode:#x}")
            if payload:
                chexdump(payload, print_fn=self.log)
            if cmd.txbuf:
                self.log(f"TX buf @ {cmd.txbuf:#x} ({cmd.txlen:#x} bytes):")
                chexdump(self.dart.ioread(self.stream, cmd.txbuf, cmd.txlen), print_fn=self.log)

KNOWN_MSGS = {
    "A000": "IOMFB::UPPipeAP_H13P::late_init_signal()",
    "A001": "IOMFB::UPPipeAP_H13P::init_ipa(unsigned long long, unsigned long)",
    "A002": "IOMFB::UPPipeAP_H13P::alss_supported()",
    "A003": "IOMFB::UPPipeAP_H13P::reset_dsc()",
    "A004": "IOMFB::UPPipeAP_H13P::display_edr_factor_changed(float)",
    "A005": "IOMFB::UPPipeAP_H13P::set_contrast(float)",
    "A006": "IOMFB::UPPipeAP_H13P::set_csc_mode(IOMFB_CSCMode)",
    "A007": "IOMFB::UPPipeAP_H13P::set_op_mode(IOMFB_CSCMode)",
    "A008": "IOMFB::UPPipeAP_H13P::set_op_gamma_mode(IOMFB_TFMode)",
    "A009": "IOMFB::UPPipeAP_H13P::set_video_out_mode(IOMFB_Output_Mode)",
    "A010": "IOMFB::UPPipeAP_H13P::set_meta_allowed(bool)",
    "A011": "IOMFB::UPPipeAP_H13P::set_tunneled_color_mode(bool)",
    "A012": "IOMFB::UPPipeAP_H13P::set_bwr_line_time_us(double)",
    "A013": "IOMFB::UPPipeAP_H13P::performance_feedback(double)",
    "A014": "IOMFB::UPPipeAP_H13P::notify_swap_complete(unsigned int)",
    "A015": "IOMFB::UPPipeAP_H13P::is_run_mode_change_pending() const",
    "A016": "IOMFB::UPPipeAP_H13P::ready_for_run_mode_change(IOMFB::AppleRegisterStream*)",
    "A017": "IOMFB::UPPipeAP_H13P::set_thermal_throttle_cap(unsigned int)",
    "A018": "IOMFB::UPPipeAP_H13P::emergency_shutdown_normal_mode(IOMFB::AppleRegisterStream*)",
    "A019": "IOMFB::UPPipeAP_H13P::set_target_run_mode(IOMFB::AppleRegisterStream*)",
    "A020": "IOMFB::UPPipeAP_H13P::rt_bandwidth_setup()",
    "A021": "IOMFB::UPPipeAP_H13P::rt_bandwidth_update(IOMFB::AppleRegisterStream*, float, float, bool, bool)",
    "A022": "IOMFB::UPPipeAP_H13P::rt_bandwidth_update_downgrade(IOMFB::AppleRegisterStream*)",
    "A023": "IOMFB::UPPipeAP_H13P::rt_bandwidth_write_update(IOMFB::AppleRegisterStream*, RealtimeBandwithWritebackBlock, bool)",
    "A024": "IOMFB::UPPipeAP_H13P::cif_blending_eco_present()",
    "A025": "IOMFB::UPPipeAP_H13P::request_bic_update()",
    "A026": "IOMFB::UPPipeAP_H13P::early_power_off_warning(IOMFB::AppleRegisterStream*)",
    "A027": "IOMFB::UPPipeAP_H13P::get_max_frame_size(unsigned int*, unsigned int*)",
    "A028": "IOMFB::UPPipeAP_H13P::shadow_FIFO_empty(IOMFB::AppleRegisterStream*) const",
    "A029": "IOMFB::UPPipeAP_H13P::setup_video_limits()",
    "A030": "IOMFB::UPPipeAP_H13P::can_program_swap() const",
    "A031": "IOMFB::UPPipeAP_H13P::in_auto_mode() const",
    "A032": "IOMFB::UPPipeAP_H13P::push_black_frame(IOMFB::AppleRegisterStream*)",
    "A033": "IOMFB::UPPipeAP_H13P::read_crc(Notify_Info_Index, unsigned int)",
    "A034": "IOMFB::UPPipeAP_H13P::update_notify_clients_dcp(unsigned int const*)",
    "A035": "IOMFB::UPPipeAP_H13P::is_hilo() const",
    "A036": "IOMFB::UPPipeAP_H13P::apt_supported() const",
    "A037": "IOMFB::UPPipeAP_H13P::get_dfb_info(unsigned int*, unsigned long long*, unsigned int*)",
    "A038": "IOMFB::UPPipeAP_H13P::get_dfb_compression_info(unsigned int*)",
    "A039": "IOMFB::UPPipeAP_H13P::get_frame_done_time() const",
    "A040": "IOMFB::UPPipeAP_H13P::get_performance_headroom() const",
    "A041": "IOMFB::UPPipeAP_H13P::are_stats_active() const",
    "A042": "IOMFB::UPPipeAP_H13P::supports_odd_h_blanking() const",
    "A043": "IOMFB::UPPipeAP_H13P::is_first_hw_version() const",
    "A044": "IOMFB::UPPipeAP_H13P::set_blendout_CSC_mode()",

    "A100": "IOMFB::UPPipe2::get_gamma_table_gated(IOMFBGammaTable*)",
    "A101": "IOMFB::UPPipe2::set_gamma_table_gated(IOMFBGammaTable const*)",
    "A102": "IOMFB::UPPipe2::test_control(IOMFB_TC_Cmd, unsigned int)",
    "A103": "IOMFB::UPPipe2::get_config_frame_size(unsigned int*, unsigned int*) const",
    "A104": "IOMFB::UPPipe2::set_config_frame_size(unsigned int, unsigned int) const",
    "A105": "IOMFB::UPPipe2::program_config_frame_size() const",
    "A106": "IOMFB::UPPipe2::read_blend_crc() const",
    "A107": "IOMFB::UPPipe2::read_config_crc() const",
    "A108": "IOMFB::UPPipe2::disable_wpc_calibration(bool)",
    "A109": "IOMFB::UPPipe2::vftg_is_running(IOMFB::AppleRegisterStream*) const",
    "A110": "IOMFB::UPPipe2::vftg_debug(IOMFB::AppleRegisterStream*, unsigned int) const",
    "A111": "IOMFB::UPPipe2::vftg_set_color_channels(unsigned int, unsigned int, unsigned int)",
    "A112": "IOMFB::UPPipe2::set_color_filter_scale(int)",
    "A113": "IOMFB::UPPipe2::set_corner_temps(int const*)",
    "A114": "IOMFB::UPPipe2::reset_aot_enabled() const",
    "A115": "IOMFB::UPPipe2::aot_enabled() const",
    "A116": "IOMFB::UPPipe2::aot_active() const",
    "A117": "IOMFB::UPPipe2::set_timings_enabled(IOMFB::AppleRegisterStream*, bool)",
    "A118": "IOMFB::UPPipe2::get_frame_size(IOMFB::AppleRegisterStream*, unsigned int*, unsigned int*)",
    "A119": "IOMFB::UPPipe2::set_block(unsigned long long, unsigned int, unsigned int, unsigned long long const*, unsigned int, unsigned char const*, unsigned long, bool)",
    "A121": "IOMFB::UPPipe2::get_buf_block(unsigned long long, unsigned int, unsigned int, unsigned long long const*, unsigned int, unsigned char const*, unsigned long, bool)",
    "A122": "IOMFB::UPPipe2::get_matrix(IOMFB_MatrixLocation, IOMFBColorFixedMatrix*) const",
    "A123": "IOMFB::UPPipe2::set_matrix(IOMFB_MatrixLocation, IOMFBColorFixedMatrix const*)",
    "A124": "IOMFB::UPPipe2::get_internal_timing_attributes_gated(IOMFB::RefreshTimingAttributes*) const",
    "A125": "IOMFB::UPPipe2::display_edr_factor_changed(float)",
    "A126": "IOMFB::UPPipe2::set_contrast(float)",
    "A127": "IOMFB::UPPipe2::p3_to_disp_cs(float const*, float const (*) [2])",
    "A128": "IOMFB::UPPipe2::max_panel_brightness() const",
    "A129": "IOMFB::UPPipe2::swap_flush_stream_replay(IOMFB::AppleRegisterStream*)",
    "A130": "IOMFB::UPPipe2::init_ca_pmu()",
    "A131": "IOMFB::UPPipe2::pmu_service_matched()",
    "A132": "IOMFB::UPPipe2::backlight_service_matched()",

    "A200": "IOMFB::PropRelay::setBool(IOMFB::RuntimeProperty, bool)",
    "A201": "IOMFB::PropRelay::setInt(IOMFB::RuntimeProperty, unsigned int)",
    "A202": "IOMFB::PropRelay::setFx(IOMFB::RuntimeProperty, int)",
    "A203": "IOMFB::PropRelay::setPropDynamic(IOMFB::RuntimeProperty, unsigned int)",
    "A204": "IOMFB::PropRelay::getBool(IOMFB::RuntimeProperty)",
    "A205": "IOMFB::PropRelay::getInt(IOMFB::RuntimeProperty)",
    "A206": "IOMFB::PropRelay::getFx(IOMFB::RuntimeProperty)",

    "A350": "UnifiedPipeline2::displayHeight()",
    "A351": "UnifiedPipeline2::displayWidth()",
    "A352": "UnifiedPipeline2::applyProperty(unsigned int, unsigned int)",
    "A353": "UnifiedPipeline2::get_system_type() const",
    "A354": "UnifiedPipeline2::headless() const",
    "A355": "UnifiedPipeline2::export_idle_method(unsigned int)",
    "A357": "UnifiedPipeline2::set_create_DFB()",
    "A358": "UnifiedPipeline2::vi_set_temperature_hint()",

    "A400": "IOMobileFramebufferAP::free_signal()",
    "A401": "IOMobileFramebufferAP::start_signal()",
    "A402": "IOMobileFramebufferAP::stop_signal()",
    "A403": "IOMobileFramebufferAP::systemWillShutdown()",
    "A404": "IOMobileFramebufferAP::swap_begin()",
    "A405": "IOMobileFramebufferAP::rotate_surface(unsigned int, unsigned int, unsigned int)",
    "A406": "IOMobileFramebufferAP::get_framebuffer_id()",
    "A407": "IOMobileFramebufferAP::swap_start(unsigned int*, IOUserClient*)",
    "A408": "IOMobileFramebufferAP::swap_submit_dcp(IOMFBSwapRec const*, IOSurface**, unsigned int const*, bool, double, unsigned int, bool*)",
    "A409": "IOMobileFramebufferAP::swap_signal(unsigned int, unsigned int)",
    "A410": "IOMobileFramebufferAP::set_display_device(unsigned int)",
    "A411": "IOMobileFramebufferAP::is_main_display() const",
    "A412": "IOMobileFramebufferAP::set_digital_out_mode(unsigned int, unsigned int)",
    "A413": "IOMobileFramebufferAP::get_digital_out_state(unsigned int*)",
    "A414": "IOMobileFramebufferAP::get_display_area(DisplayArea*)",
    "A415": "IOMobileFramebufferAP::set_tvout_mode(unsigned int)",
    "A416": "IOMobileFramebufferAP::set_tvout_signaltype(unsigned int)",
    "A417": "IOMobileFramebufferAP::set_wss_info(unsigned int, unsigned int)",
    "A418": "IOMobileFramebufferAP::set_content_flags(unsigned int)",
    "A419": "IOMobileFramebufferAP::get_gamma_table(IOMFBGammaTable*)",
    "A420": "IOMobileFramebufferAP::set_gamma_table(IOMFBGammaTable*)",
    "A421": "IOMobileFramebufferAP::get_matrix(unsigned int, unsigned long long (*) [3][3]) const",
    "A422": "IOMobileFramebufferAP::set_matrix(unsigned int, unsigned long long const (*) [3][3])",
    "A423": "IOMobileFramebufferAP::set_contrast(float*)",
    "A424": "IOMobileFramebufferAP::set_white_on_black_mode(unsigned int)",
    "A425": "IOMobileFramebufferAP::set_color_remap_mode(DisplayColorRemapMode)",
    "A426": "IOMobileFramebufferAP::get_color_remap_mode(DisplayColorRemapMode*) const",
    "A427": "IOMobileFramebufferAP::setBrightnessCorrection(unsigned int)",
    "A428": "IOMobileFramebufferAP::temp_queue_swap_cancel(unsigned int)",
    "A429": "IOMobileFramebufferAP::swap_cancel(unsigned int)",
    "A430": "IOMobileFramebufferAP::swap_cancel_all_dcp(unsigned long long)",
    "A431": "IOMobileFramebufferAP::surface_is_replaceable(unsigned int, bool*)",
    "A432": "IOMobileFramebufferAP::kernel_tests(IOMFBKernelTestsArguments*)",
    "A433": "IOMobileFramebufferAP::splc_set_brightness(unsigned int)",
    "A434": "IOMobileFramebufferAP::splc_get_brightness(unsigned int*)",
    "A435": "IOMobileFramebufferAP::set_block_dcp(task*, unsigned int, unsigned int, unsigned long long const*, unsigned int, unsigned char const*, unsigned long)",
    "A436": "IOMobileFramebufferAP::get_block_dcp(task*, unsigned int, unsigned int, unsigned long long const*, unsigned int, unsigned char*, unsigned long) const",
    "A438": "IOMobileFramebufferAP::swap_set_color_matrix(IOMFBColorFixedMatrix*, IOMFBColorMatrixFunction, unsigned int)",
    "A439": "IOMobileFramebufferAP::set_parameter_dcp(IOMFBParameterName, unsigned long long const*, unsigned int)",
    "A440": "IOMobileFramebufferAP::display_width() const",
    "A441": "IOMobileFramebufferAP::display_height() const",
    "A442": "IOMobileFramebufferAP::get_display_size(unsigned int*, unsigned int*) const",
    "A443": "IOMobileFramebufferAP::do_create_default_frame_buffer() const",
    "A444": "IOMobileFramebufferAP::printRegs()",
    "A445": "IOMobileFramebufferAP::enable_disable_dithering(unsigned int)",
    "A446": "IOMobileFramebufferAP::set_underrun_color(unsigned int)",
    "A447": "IOMobileFramebufferAP::enable_disable_video_power_savings(unsigned int)",
    "A448": "IOMobileFramebufferAP::set_video_dac_gain(unsigned int)",
    "A449": "IOMobileFramebufferAP::set_line21_data(unsigned int)",
    "A450": "IOMobileFramebufferAP::enableInternalToExternalMirroring(bool)",
    "A451": "IOMobileFramebufferAP::getExternalMirroringCapability(IOMFBMirroringCapability*)",
    "A452": "IOMobileFramebufferAP::setRenderingAngle(float*)",
    "A453": "IOMobileFramebufferAP::setOverscanSafeRegion(IOMFBOverscanSafeRect*)",
    "A454": "IOMobileFramebufferAP::first_client_open()",
    "A455": "IOMobileFramebufferAP::last_client_close_dcp(unsigned int*)",
    "A456": "IOMobileFramebufferAP::writeDebugInfo(unsigned long)",
    "A457": "IOMobileFramebufferAP::flush_debug_flags(unsigned int)",
    "A458": "IOMobileFramebufferAP::io_fence_notify(unsigned int, unsigned int, unsigned long long, IOMFBStatus)",
    "A459": "IOMobileFramebufferAP::swap_wait_dcp(bool, unsigned int, unsigned int, unsigned int)",
    "A460": "IOMobileFramebufferAP::setDisplayRefreshProperties()",
    "A461": "IOMobileFramebufferAP::exportProperty(unsigned int, unsigned int)",
    "A462": "IOMobileFramebufferAP::applyProperty(unsigned int, unsigned int)",
    "A463": "IOMobileFramebufferAP::flush_supportsPower(bool)",
    "A464": "IOMobileFramebufferAP::abort_swaps_dcp(IOMobileFramebufferUserClient*)",
    "A465": "IOMobileFramebufferAP::swap_signal_gated(unsigned int, unsigned int)",
    "A466": "IOMobileFramebufferAP::update_dfb(IOSurface*, unsigned int, unsigned int, unsigned long long)",
    "A467": "IOMobileFramebufferAP::update_dfb(IOSurface*)",
    "A468": "IOMobileFramebufferAP::setPowerState(unsigned long, bool, unsigned int*)",
    "A469": "IOMobileFramebufferAP::isKeepOnScreen() const",
    "A470": "IOMobileFramebufferAP::resetStats()",
    "A471": "IOMobileFramebufferAP::set_has_frame_swap_function(bool)",
    "A472": "IOMobileFramebufferAP::getPerformanceStats(unsigned int*, unsigned int*)",

    "D000": "bool IOMFB::UPPipeAP_H13P::did_boot_signal()",
    "D001": "bool IOMFB::UPPipeAP_H13P::did_power_on_signal()",
    "D002": "void IOMFB::UPPipeAP_H13P::will_power_off_signal()",
    "D003": "void IOMFB::UPPipeAP_H13P::rt_bandwidth_setup_ap(inout rt_bw_config_t*)",
    "D004": "void IOMFB::UPPipeAP_H13P::mcc_report_replay(bool, unsigned int)",
    "D005": "void IOMFB::UPPipeAP_H13P::mcc_report_bics(bool, unsigned int)",

    "D100": "void UnifiedPipeline2::match_pmu_service()",
    #"D101": "", # get some uint32_t, inlined
    "D102": "void UnifiedPipeline2::set_number_property(char const*, unsigned int)",
    "D103": "void UnifiedPipeline2::set_boolean_property(char const*, bool)",
    "D104": "void UnifiedPipeline2::set_string_property(char const*, char const*)",
    "D105": "IOReturn IOService::acknowledgeSetPowerState()",
    "D106": "void IORegistryEntry::removeProperty(char const*)",
    "D107": "bool UnifiedPipeline2::create_provider_service",
    "D108": "bool UnifiedPipeline2::create_product_service()",
    "D109": "bool UnifiedPipeline2::create_PMU_service()",
    "D110": "bool UnifiedPipeline2::create_iomfb_service()",
    "D111": "bool UnifiedPipeline2::create_backlight_service()",
    "D112": "void UnifiedPipeline2::set_idle_caching_state_ap(IdleCachingState, unsigned int)",
    "D113": "bool UnifiedPipeline2::upload_trace_start(IOMFB::FrameInfoBuffer::FrameInfoConstants const*)",
    "D114": "bool UnifiedPipeline2::upload_trace_chunk(IOMFB::FrameInfoBuffer::FrameInfoData const*, unsigned int, unsigned int)",
    "D115": "bool UnifiedPipeline2::upload_trace_end(char const*)",
    "D116": "bool UnifiedPipeline2::start_hardware_boot()",
    "D117": "bool UnifiedPipeline2::is_dark_boot()",
    "D118": "bool UnifiedPipeline2::is_waking_from_hibernate()",
    "D119": "bool UnifiedPipeline2::detect_fastsim()",
    "D120": "bool UnifiedPipeline2::read_edt_data(char const*, unsigned int, unsigned int*) const",
    "D121": "bool UnifiedPipeline2::read_edt_string(char const*, char*, unsigned int*) const",
    "D122": "bool UnifiedPipeline2::setDCPAVPropStart(unsigned int)",
    "D123": "bool UnifiedPipeline2::setDCPAVPropChunk(unsigned char const*, unsigned int, unsigned int)",
    "D124": "bool UnifiedPipeline2::setDCPAVPropEnd(char const*)",

    "D200": "uint64_t IOMFB::UPPipe2::get_default_idle_caching_method()",
    "D201": "IOMFBStatus IOMFB::UPPipe2::map_buf(IOMFB::BufferDescriptor*, unsigned long*, unsigned long long*, bool)",
    "D202": "void IOMFB::UPPipe2::unmap_buf(IOMFB::BufferDescriptor*, unsigned long, unsigned long long, bool)",
    "D203": "bool IOMFB::UPPipe2::aot_supported_peek()",
    "D204": "uint64_t IOMFB::UPPipe2::get_ideal_screen_space()",
    "D205": "bool IOMFB::UPPipe2::read_carveout(unsigned char*, unsigned int, unsigned int) const",
    "D206": "bool IOMFB::UPPipe2::match_pmu_service()",
    "D207": "bool IOMFB::UPPipe2::match_backlight_service()",
    "D208": "uint64_ IOMFB::UPPipe2::get_calendar_time_ms()",
    "D209": "void IOMFB::UPPipe2::plc_enable(bool)",
    "D210": "void IOMFB::UPPipe2::plc_init()",
    "D211": "void IOMFB::UPPipe2::update_backlight_factor_prop(int)",

    "D300": "void IOMFB::PropRelay::publish(IOMFB::RuntimeProperty, unsigned int)",

    "D400": "void IOMFB::ServiceRelay::get_property(unsigned int, in char const[0x40], out unsigned char[0x200], inout unsigned int*)",
    "D401": "bool IOMFB::ServiceRelay::get_uint_prop(unsigned int, in char const[0x40], inout unsigned long long*)",
    "D402": "void IOMFB::ServiceRelay::set_uint_prop(unsigned int, in char const[0x40], unsigned long long)",
    "D403": "bool IOMFB::ServiceRelay::get_uint_prop(unsigned int, in char const[0x40], inout unsigned int*)",
    "D404": "void IOMFB::ServiceRelay::set_uint_prop(unsigned int, in char const[0x40], unsigned int)",
    "D405": "bool IOMFB::ServiceRelay::get_fx_prop(unsigned int, in char const[0x40], inout int*)",
    "D406": "void IOMFB::ServiceRelay::set_fx_prop(unsigned int, in char const[0x40], int)",
    "D407": "void IOMFB::ServiceRelay::set_bool_prop(unsigned int, in char const[0x40], bool)",
    "D408": "unsigned long long IOMFB::ServiceRelay::getClockFrequency(unsigned int, unsigned int)",
    "D409": "IOMFBStatus IOMFB::ServiceRelay::enableDeviceClock(unsigned int, unsigned int, unsigned int)",
    "D410": "IOMFBStatus IOMFB::ServiceRelay::enableDevicePower(unsigned int, unsigned int, inout unsigned int*, unsigned int)",
    "D411": "IOMFBStatus IOMFB::ServiceRelay::mapDeviceMemoryWithIndex(unsigned int, unsigned int, unsigned int, inout unsigned long*, inout unsigned long long*)",
    "D412": "bool IOMFB::ServiceRelay::setProperty(unsigned int, OSString<0x40> const*, OSArray const*)",
    "D413": "bool IOMFB::ServiceRelay::setProperty(unsigned int, OSString<0x40> const*, OSDictionary const*)",
    "D414": "bool IOMFB::ServiceRelay::setProperty(unsigned int, OSString<0x40> const*, OSNumber const*)",
    "D415": "bool IOMFB::ServiceRelay::setProperty(unsigned int, OSString<0x40> const*, OSBoolean const*)",
    "D416": "bool IOMFB::ServiceRelay::setProperty(unsigned int, OSString<0x40> const*, OSString const*)",
    "D417": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], OSArray const*)",
    "D418": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], OSDictionary const*)",
    "D419": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], OSNumber const*)",
    "D420": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], OSBoolean const*)",
    "D421": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], OSString const*)",
    "D422": "bool IOMFB::ServiceRelay::setProperty(unsigned int, char const[0x40], bool)",
    "D423": "void IOMFB::ServiceRelay::removeProperty(unsigned int, char const[0x40])",
    "D424": "void IOMFB::ServiceRelay::removeProperty(unsigned int, OSString<0x40> const*)",

    "D450": "bool IOMFB::MemDescRelay::from_id(unsigned int, unsigned long*, unsigned long*, unsigned long long*)",
    "D451": "MemDescRelay::desc_id_t IOMFB::MemDescRelay::allocate_buffer(unsigned int, unsigned long long, unsigned int, unsigned long*, unsigned long*, unsigned long long*)",
    "D452": "MemDescRelay::desc_id_t IOMFB::MemDescRelay::map_physical(unsigned long long, unsigned long long, unsigned int, unsigned long*, unsigned long long*)",
    "D453": "MemDescRelay::desc_id_t IOMFB::MemDescRelay::withAddressRange(unsigned long long, unsigned long long, unsigned int, task*, unsigned long*, unsigned long long*)",
    "D454": "IOMFBStatus IOMFB::MemDescRelay::prepare(unsigned int, unsigned int)",
    "D455": "IOMFBStatus IOMFB::MemDescRelay::complete(unsigned int, unsigned int)",
    "D456": "bool IOMFB::MemDescRelay::release_descriptor(unsigned int)",

    "D500": "IOMFBStatus IOMFB::PlatformFunctionRelay::allocate_record(unsigned int, char const*, unsigned int, bool)",
    "D501": "IOMFBStatus IOMFB::PlatformFunctionRelay::release_record(unsigned int)",
    "D502": "IOMFBStatus IOMFB::PlatformFunctionRelay::callFunctionLink(unsigned int, unsigned long, unsigned long, unsigned long)",

    "D550": "bool IORegistryEntry::setProperty(OSString *, OSArray *)",
    "D551": "bool IORegistryEntry::setProperty(OSString *, IOMFB::AFKArray *)",
    "D552": "bool IORegistryEntry::setProperty(OSString *, OSDictionary *)",
    "D553": "bool IORegistryEntry::setProperty(OSString *, IOMFB::AFKDictionary *)",
    "D554": "bool IORegistryEntry::setProperty(OSString *, OSNumber *)",
    "D555": "bool IORegistryEntry::setProperty(OSString *, IOMFB::AFKNumber *)",
    "D556": "bool IORegistryEntry::setProperty(OSString *, OSBoolean *)",
    "D557": "bool IORegistryEntry::setProperty(OSString *, OSString *)",
    "D558": "bool IORegistryEntry::setProperty(OSString *, IOMFB::AFKString *)",
    "D559": "bool IORegistryEntry::setProperty(char const*, OSArray *)",
    "D560": "bool IORegistryEntry::setProperty(char const*, IOMFB::AFKArray *)",
    "D561": "bool IORegistryEntry::setProperty(char const*, OSDictionary *)",
    "D562": "bool IORegistryEntry::setProperty(char const*, IOMFB::AFKDictionary *)",
    "D563": "bool IORegistryEntry::setProperty(char const*, OSNumber *)",
    "D564": "bool IORegistryEntry::setProperty(char const*, IOMFB::AFKNumber *)",
    "D565": "bool IORegistryEntry::setProperty(char const*, OSBoolean *)",
    "D566": "bool IORegistryEntry::setProperty(char const*, OSString *)",
    "D567": "bool IORegistryEntry::setProperty(char const*, IOMFB::AFKString *)",
    "D568": "bool IORegistryEntry::setProperty(char const*, char const*)",
    "D569": "bool IORegistryEntry::setProperty(char const*, bool)",
    "D570": "IOMFBStatus IOMobileFramebufferAP::setProperties(OSDictionary*)",
    "D571": "void IOMobileFramebufferAP::swapping_client_did_start(IOMobileFramebufferUserClient*)",
    "D572": "void IOMobileFramebufferAP::swapping_client_will_stop(IOMobileFramebufferUserClient*)",
    "D573": "IOMFBStatus IOMobileFramebufferAP::set_canvas_size(unsigned int, unsigned int)",
    "D574": "IOMFBStatus IOMobileFramebufferAP::powerUpDART(bool)",
    "D575": "IOMFBStatus IOMobileFramebufferAP::get_dot_pitch(unsigned int*)",
    "D576": "void IOMobileFramebufferAP::hotPlug_notify_gated(unsigned long long)",
    "D577": "void IOMobileFramebufferAP::powerstate_notify(bool, bool)",
    "D578": "bool IOMobileFramebufferAP::idle_fence_create(IdleCachingState)",
    "D579": "void IOMobileFramebufferAP::idle_fence_complete()",
    "D580": "void IOMobileFramebufferAP::idle_surface_release_ap()",
    "D581": "void IOMobileFramebufferAP::swap_complete_head_of_line(unsigned int, bool, unsigned int, bool)",
    "D582": "bool IOMobileFramebufferAP::create_default_fb_surface(unsigned int, unsigned int)",
    "D583": "bool IOMobileFramebufferAP::serializeDebugInfoCb(unsigned long, IOMFB::BufferDescriptor const*, unsigned int)",
    "D584": "void IOMobileFramebufferAP::clear_default_surface()",
    "D585": "void IOMobileFramebufferAP::swap_notify_gated(unsigned long long, unsigned long long, unsigned long long)",
    "D586": "void IOMobileFramebufferAP::swap_info_notify_dispatch(SwapInfoBlob const*)",
    "D587": "void IOMFBStatus IOMobileFramebufferAP::updateBufferMappingCount_gated(bool)",
    "D588": "void IOMobileFramebufferAP::resize_default_fb_surface_gated()",
    "D589": "void IOMobileFramebufferAP::swap_complete_ap_gated(unsigned int, bool, SwapCompleteData const*, SwapInfoBlob const*, unsigned int)",
    "D590": "void IOMobileFramebufferAP::batched_swap_complete_ap_gated(unsigned int*, unsigned int, bool, bool const*, SwapCompleteData const*)",
    "D591": "void IOMobileFramebufferAP::swap_complete_intent_gated(unsigned int, bool, IOMFBSwapRequestIntent, unsigned int, unsigned int)",
    "D592": "void IOMobileFramebufferAP::abort_swap_ap_gated(unsigned int)",
    "D593": "void IOMobileFramebufferAP::enable_backlight_message_ap_gated(bool)",
    "D594": "void IOMobileFramebufferAP::setSystemConsoleMode(bool)",
    "D595": "void IOMobileFramebufferAP::setSystemConsoleMode_gated(bool)",
    "D596": "bool IOMobileFramebufferAP::isDFBAllocated()",
    "D597": "bool IOMobileFramebufferAP::preserveContents()",
    "D598": "void IOMobileFramebufferAP::find_swap_function_gated()",

    "D700": "int IOMFB::DCPPowerManager::set_kernel_power_assert(bool, bool)",
}

# iboot interface
"""
0: setResource
1: setSurface
2: setPower
3: getHpdStatus
4: getTimingModes
5: getColorModes
6: setMode
7: setBrightness
8: rwBCONRegsRequest
9: setParameter
10: setMatrix
11: setProperty
12: getProperty
13: setBlock
14: getBlock
15: swapBegin
16: setSwapLayer
17: setSwapTimestamp
18: setSwapEnd
19: setSwapWait
20: setBrightnessCfg
21: setNamedProperty
22: getNamedProperty
"""

from m1n1.fw.dcp.dcpep import DCPMessage, DCPEp_SetShmem, CallContext, DCPEp_Msg

class DCPCallState:
    pass

class DCPCallChannel(Reloadable):
    def __init__(self, dcpep, name, buf, bufsize):
        self.dcpep = dcpep
        self.name = name
        self.buf = buf
        self.bufsize = bufsize
        self.log = self.dcpep.log
        self.state = self.dcpep.state

    def call(self, msg, dir):
        ident = f"{self.name}.{msg.OFF:x}"

        if any(msg.OFF == s.off for s in self.state.ch.get(self.name, [])):
            self.log(f"{dir}{self.name}.{msg.OFF:x} !!! Overlapping call ({msg})")
            assert False

        state = DCPCallState()

        data = self.dcpep.dart.ioread(self.dcpep.stream, self.state.shmem_iova + self.buf + msg.OFF, msg.LEN)
        tag = data[:4][::-1].decode("ascii")
        in_len, out_len = struct.unpack("<II", data[4:12])
        data_in = data[12:12 + in_len]

        state.off = msg.OFF
        state.tag = tag
        state.in_len = in_len
        state.out_addr = self.buf + msg.OFF + 12 + in_len
        state.out_len = out_len

        verb = self.dcpep.get_verbosity(tag)
        if verb >= 1:
            self.log(f"{dir}{self.name}.{msg.OFF:x} {tag}:{KNOWN_MSGS.get(tag, 'unk')} ({msg})")
        if verb >= 2:
            print(f"Message: {tag} ({KNOWN_MSGS.get(tag, 'unk')}): (in {in_len:#x}, out {out_len:#x})")
            if data_in:
                print(f"{dir} Input ({len(data_in):#x} bytes):")
                chexdump(data_in[:self.state.max_len])

        #if tag not in KNOWN_MSGS:
            #hv.run_shell()

        if self.state.dumpfile:
            dump = f"CALL {dir} {msg.value:#018x} {self.name} {state.off:#x} {state.tag} {in_len:#x} {out_len:#x} {data_in.hex()}\n"
            self.state.dumpfile.write(dump)
            self.state.dumpfile.flush()

        self.state.ch.setdefault(self.name, []).append(state)

    def ack(self, msg, dir):
        assert msg.LEN == 0

        states = self.state.ch.get(self.name, None)
        if not states:
            self.log(f"{dir}ACK {self.name}.{msg.OFF:x} !!! ACK without call ({msg})")
            return

        state = states[-1]

        if self.state.show_acks:
            self.log(f"{dir}ACK {self.name}.{msg.OFF:x} ({msg})")

        data_out = self.dcpep.dart.ioread(self.dcpep.stream, self.state.shmem_iova + state.out_addr, state.out_len)

        verb = self.dcpep.get_verbosity(state.tag)
        if verb >= 3 and state.out_len > 0:
            print(f"{dir}{self.name}.{msg.OFF:x} Output buffer ({len(data_out):#x} bytes):")
            chexdump(data_out[:self.state.max_len])

        if self.state.dumpfile:
            dump = f"ACK {dir} {msg.value:#018x} {self.name} {state.off:#x} {data_out.hex()}\n"
            self.state.dumpfile.write(dump)
            self.state.dumpfile.flush()

        states.pop()

class DCPEp(EP):
    BASE_MESSAGE = DCPMessage

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.state.shmem_iova = None
        self.state.show_globals = True
        self.state.show_acks = True
        self.state.max_len = 1024 * 1024
        self.state.verbosity = 3
        self.state.op_verb = {}
        self.state.ch = {}
        self.state.dumpfile = None

        self.ch_cmd      = DCPCallChannel(self, "CMD",      0x00000, 0x8000)
        self.ch_oobcmd   = DCPCallChannel(self, "OOBCMD",   0x08000, 0x8000)
        self.ch_async    = DCPCallChannel(self, "ASYNC",    0x40000, 0x8000)
        self.ch_oobasync = DCPCallChannel(self, "OOBASYNC", 0x48000, 0x8000)
        self.ch_cb       = DCPCallChannel(self, "CB",       0x60000, 0x8000)
        self.ch_oobcb    = DCPCallChannel(self, "OOBCB",    0x68000, 0x8000)

        self.cmd_ch = {
            CallContext.CB: self.ch_cmd,
            CallContext.CMD: self.ch_cmd,
            CallContext.ASYNC: None, # unknown
            CallContext.OOBASYNC: None, # unknown
            CallContext.OOBCB: self.ch_oobcmd,
            CallContext.OOBCMD: self.ch_oobcmd,
        }

        self.cb_ch = {
            CallContext.CB: self.ch_cb,
            CallContext.CMD: None,
            CallContext.ASYNC: self.ch_async,
            CallContext.OOBASYNC: self.ch_oobasync,
            CallContext.OOBCB: self.ch_oobcb,
            CallContext.OOBCMD: None,
        }

    def start(self):
        self.add_mon()

    def add_mon(self):
        if self.state.shmem_iova and self.state.show_globals:
            addr = self.state.shmem_iova + 0x80000
            iomon.add(addr, 128,
                      name=f"{self.name}.shmem@{addr:08x}", offset=addr)

            #addr = self.state.shmem_iova
            #iomon.add(addr, 0x80080,
                      #name=f"{self.name}.shmem@{addr:08x}", offset=addr)

    InitComplete = msg_log(1, DIR.RX)

    @msg(0, DIR.TX, DCPEp_SetShmem)
    def SetShmem(self, msg):
        self.log(f"Shared memory DVA: {msg.DVA:#x}")
        self.state.shmem_iova = msg.DVA & 0xfffffffff
        self.add_mon()

    @msg(2, DIR.TX, DCPEp_Msg)
    def Tx(self, msg):
        if msg.ACK:
            self.cb_ch[msg.CTX].ack(msg, ">")
        else:
            self.cmd_ch[msg.CTX].call(msg, ">")

        if self.state.show_globals:
            iomon.poll()

        return True

    @msg(2, DIR.RX, DCPEp_Msg)
    def Rx(self, msg):
        self.log(msg)
        if msg.ACK:
            self.cmd_ch[msg.CTX].ack(msg, "<")
        else:
            self.cb_ch[msg.CTX].call(msg, "<")

        if self.state.show_globals:
            iomon.poll()

        return True

    def get_verbosity(self, tag):
        return self.state.op_verb.get(tag, self.state.verbosity)

    def set_verb_known(self, verb):
        for i in KNOWN_MSGS:
            if verb is None:
                self.state.op_verb.pop(i, None)
            else:
                self.state.op_verb[i] = verb

class SystemService(EPICEp):
    NAME = "system"

class TestService(EPICEp):
    NAME = "test"

class DCPExpertService(EPICEp):
    NAME = "dcpexpert"

class Disp0Service(EPICEp):
    NAME = "disp0"

class DCPAVControllerEpicTracer(EPICServiceTracer):
    NAME = "dcpav-controller-epic"

    @epic_service_cmd(0, 14)
    def getParticipatesPowerManagement(self, data):
        self.log("> getParticipatesPowerManagement")
    @epic_service_reply(0, 14)
    def getParticipatesPowerManagement_reply(self, data):
        self.log("< getParticipatesPowerManagement")
        chexdump(data, print_fn=self.log)

class DPAVController(EPICEp):
    NAME = "dpavctrl"
    SERVICES = [
        DCPAVControllerEpicTracer
    ]

class DPSACService(EPICEp):
    NAME = "dpsac"


class DCPDPDeviceEpicTracer(EPICServiceTracer):
    NAME = "dcpdp-device-epic"

    @epic_service_cmd(0, 15)
    def getDeviceMatchingData(self, data):
        self.log("> getDeviceMatchingData")
    @epic_service_reply(0, 15)
    def getDeviceMatchingData_reply(self, data):
        self.log("< getDeviceMatchingData")
        chexdump(data, print_fn=self.log)

class DPDevService(EPICEp):
    NAME = "dpdev"
    SERVICES = [
        DCPDPDeviceEpicTracer
    ]

class DPAVService(EPICEp):
    NAME = "dpavserv"

class DCPAVAudioInterfaceEpicTracer(EPICServiceTracer):
    NAME = "dcpav-audio-interface-epic"

    # usually 4, 6 but apparently also 0, 6 here?
    # or maybe a different open?
    @epic_service_cmd(0, 6)
    def open2(self, data):
        self.log("> open")
    @epic_service_reply(0, 6)
    def open2_reply(self, data):
        self.log("< open")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 8)
    def prepareLink(self, data):
        self.log("> prepareLink")
    @epic_service_reply(0, 8)
    def prepareLink_reply(self, data):
        self.log("< prepareLink")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 9)
    def startLink(self, data):
        self.log("> startLink")
    @epic_service_reply(0, 9)
    def startLink_reply(self, data):
        self.log("< startLink")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 15)
    def getLinkStatus(self, data):
        self.log("> getLinkStatus")
    @epic_service_reply(0, 15)
    def getLinkStatus_reply(self, data):
        self.log("< getLinkStatus")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 16)
    def getTransport(self, data):
        self.log("> getTransport")
    @epic_service_reply(0, 16)
    def getTransport_reply(self, data):
        self.log("< getTransport")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 17)
    def getPortID(self, data):
        self.log("> getPortID")
    @epic_service_reply(0, 17)
    def getPortID_reply(self, data):
        self.log("< getPortID")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(1, 18)
    def getElements(self, data):
        self.log("> getElements")
    @epic_service_reply(1, 18)
    def getElements_reply(self, data):
        self.log("< getElements")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(1, 20)
    def getProductAttributes(self, data):
        self.log("> getProductAttributes")
    @epic_service_reply(1, 20)
    def getProductAttributes_reply(self, data):
        self.log("< getProductAttributes")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(1, 21)
    def getEDIDUUID(self, data):
        self.log("> getEDIDUUID")
    @epic_service_reply(1, 21)
    def getEDIDUUID_reply(self, data):
        self.log("< getEDIDUUID")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 22)
    def getDataLatency(self, data):
        self.log("> getDataLatency")
    @epic_service_reply(0, 22)
    def getDataLatency_reply(self, data):
        self.log("< getDataLatency")
        chexdump(data, print_fn=self.log)


class AVService(EPICEp):
    NAME = "av"
    SERVICES = [
        DCPAVAudioInterfaceEpicTracer
    ]

class DCPDPTXHDCPAuthSessionTracer(EPICServiceTracer):
    NAME = "dcpdptx-hdcp-auth-session"

    @epic_service_cmd(4, 8)
    def getProtocol(self, data):
        self.log("> getProtocol")
    @epic_service_reply(4, 8)
    def getProtocol_reply(self, data):
        self.log("< getProtocol")
        chexdump(data, print_fn=self.log)

class HDCPService(EPICEp):
    NAME = "hdcp"
    SERVICES = [
        DCPDPTXHDCPAuthSessionTracer
    ]

class RemoteAllocService(EPICEp):
    NAME = "remotealloc"

class DCPDPTXRemotePortTarget(Register32):
    CORE = 3, 0
    ATC = 7, 4
    DIE = 11, 8
    CONNECTED = 15, 15

class DCPDPTXPortEpicTracer(EPICServiceTracer):
    NAME = "dcpdptx-port-epic"

    @epic_service_cmd(0, 6)
    def setPowerState(self, data):
        self.log("> setPowerState")
        chexdump(data, print_fn=self.log)
    @epic_service_reply(0, 6)
    def setPowerState_reply(self, data):
        self.log("< setPowerState")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 11)
    def connectTo(self, data):
        unk1, target = struct.unpack("<II24x", data)
        target = DCPDPTXRemotePortTarget(target)
        self.log(f"> connectTo(target={target}, unk1=0x{unk1:x})")
        chexdump(data, print_fn=self.log)
    @epic_service_reply(0, 11)
    def connectTo_reply(self, data):
        if len(data) < 32:
            self.log(f"< connectTo() short data len={len(data)}")
        unk1, target = struct.unpack("<II24x", data)
        target = DCPDPTXRemotePortTarget(target)
        self.log(f"< connectTo(target={target}, unk1=0x{unk1:x})")
        chexdump(data, print_fn=self.log)

    @epic_service_cmd(0, 12)
    def validateConnection(self, data):
        unk1, target = struct.unpack("<II40x", data)
        target = DCPDPTXRemotePortTarget(target)
        self.log(f"> validateConnection(target={target}, unk1=0x{unk1:x})")
    @epic_service_reply(0, 12)
    def validateConnection_reply(self, data):
        unk1, target = struct.unpack("<II40x", data)
        target = DCPDPTXRemotePortTarget(target)
        self.log(f"< validateConnection(target={target}, unk1=0x{unk1:x})")

    @epic_service_cmd(8, 8)
    def hotPlugDetectChangeOccurred(self, data):
        self.log(f"> hotPlugDetectChangeOccurred()")
        chexdump(data, print_fn=self.log)
        unk = struct.unpack("<16x?15x", data)[0]
        self.log(f"> hotPlugDetectChangeOccurred(unk={unk})")
    @epic_service_reply(8, 8)
    def hotPlugDetectChangeOccurred_reply(self, data):
        if len(data) < 32:
            self.log(f"< hotPlugDetectChangeOccurred()")
            chexdump(data, print_fn=self.log)
        else:
            unk = struct.unpack("<16x?15x", data)[0]
            self.log(f"< hotPlugDetectChangeOccurred(unk={unk})")

    @epic_service_notify(0, 0)
    def apcall_activate(self, data):
        self.log(f"< DPTX_APCALL_ACTIVATE)")
    @epic_service_notify_ack(0, 0)
    def apcall_activate_ack(self, data):
        self.log(f"> DPTX_APCALL_ACTIVATE)")

    @epic_service_notify(0, 1)
    def apcall_deactivate(self, data):
        self.log(f"< DPTX_APCALL_DEACTIVATE)")
    @epic_service_notify_ack(0, 1)
    def apcall_deactivate_ack(self, data):
        self.log(f"> DPTX_APCALL_DEACTIVATE)")

    @epic_service_notify(0, 2)
    def apcall_max_drive_settings(self, data):
        self.log(f"< DPTX_APCALL_GET_MAX_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 2)
    def apcall_max_drive_settings_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_MAX_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 3)
    def apcall_set_drive_settings(self, data):
        self.log(f"< DPTX_APCALL_SET_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 3)
    def apcall_set_drive_settings_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 4)
    def apcall_get_drive_settings(self, data):
        self.log(f"< DPTX_APCALL_GET_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 4)
    def apcall_get_drive_settings_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_DRIVE_SETTINGS)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 5)
    def apcall_will_change_link_cfg(self, data):
        self.log(f"< DPTX_APCALL_WILL_CHANGE_LINK_CONFIG)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 5)
    def apcall_will_change_link_cfg_ack(self, data):
        self.log(f"> DPTX_APCALL_WILL_CHANGE_LINK_CONFIG)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 6)
    def apcall_did_change_link_cfg(self, data):
        self.log(f"< DPTX_APCALL_DID_CHANGE_LINK_CONFIG)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 6)
    def apcall_did_change_link_cfg_ack(self, data):
        self.log(f"> DPTX_APCALL_DID_CHANGE_LINK_CONFIG)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 7)
    def apcall_max_link_rate(self, data):
        self.log(f"< DPTX_APCALL_GET_MAX_LINK_RATE)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 7)
    def apcall_max_link_rate_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_MAX_LINK_RATE)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 8)
    def apcall_get_link_rate(self, data):
        self.log(f"< DPTX_APCALL_GET_LINK_RATE)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 8)
    def apcall_get_link_rate_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_LINK_RATE)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 9)
    def apcall_set_link_rate(self, data):
        self.log(f"< DPTX_APCALL_SET_LINK_RATE)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 9)
    def apcall_set_link_rate_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_LINK_RATE)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 10)
    def apcall_max_lane_count(self, data):
        self.log(f"< DPTX_APCALL_GET_MAX_LANE_COUNT)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 10)
    def apcall_max_lane_count_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_MAX_LANE_COUNT)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 11)
    def apcall_get_active_lane_count(self, data):
        self.log(f"< DPTX_APCALL_GET_ACTIVE_LANE_COUNT)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 11)
    def apcall_get_active_lane_count_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_ACTIVE_LANE_COUNT)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 12)
    def apcall_set_active_lane_count(self, data):
        self.log(f"< DPTX_APCALL_SET_ACTIVE_LANE_COUNT)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 12)
    def apcall_set_active_lane_count_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_ACTIVE_LANE_COUNT)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 13)
    def apcall_supports_downspread(self, data):
        self.log(f"< DPTX_APCALL_GET_SUPPORTS_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 13)
    def apcall_supports_downspread_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_SUPPORTS_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 14)
    def apcall_get_downspread(self, data):
        self.log(f"< DPTX_APCALL_GET_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 14)
    def apcall_get_downspread_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 15)
    def apcall_set_downspread(self, data):
        self.log(f"< DPTX_APCALL_SET_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 15)
    def apcall_set_downspread_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_DOWN_SPREAD)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 16)
    def apcall_supports_lane_map(self, data):
        self.log(f"< DPTX_APCALL_GET_SUPPORTS_LANE_MAPPING)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 16)
    def apcall_supports_lane_map_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_SUPPORTS_LANE_MAPPING)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 17)
    def apcall_set_lane_map(self, data):
        self.log(f"< DPTX_APCALL_SET_LANE_MAP)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 17)
    def apcall_set_lane_map_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_LANE_MAP)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 18)
    def apcall_supports_hpd(self, data):
        self.log(f"< DPTX_APCALL_GET_SUPPORTS_HPD)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 18)
    def apcall_supports_hpd_ack(self, data):
        self.log(f"> DPTX_APCALL_GET_SUPPORTS_HPD)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 19)
    def apcall_force_hpd(self, data):
        self.log(f"< DPTX_APCALL_FORCE_HOTPLUG_DETECT)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 19)
    def apcall_force_hpd_ack(self, data):
        self.log(f"> DPTX_APCALL_FORCE_HOTPLUG_DETECT)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 20)
    def apcall_inactive_sink(self, data):
        self.log(f"< DPTX_APCALL_INACTIVE_SINK_DETECTED)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 20)
    def apcall_inactive_sink_ack(self, data):
        self.log(f"> DPTX_APCALL_INACTIVE_SINK_DETECTED)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 21)
    def apcall_set_tiled_hints(self, data):
        self.log(f"< DPTX_APCALL_SET_TILED_DISPLAY_HINTS)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 21)
    def apcall_set_tiled_hints_ack(self, data):
        self.log(f"> DPTX_APCALL_SET_TILED_DISPLAY_HINTS)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 22)
    def apcall_dev_not_responding(self, data):
        self.log(f"< DPTX_APCALL_DEVICE_NOT_RESPONDING)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 22)
    def apcall_dev_not_responding_ack(self, data):
        self.log(f"> DPTX_APCALL_DEVICE_NOT_RESPONDING)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 23)
    def apcall_dev_busy_timeout(self, data):
        self.log(f"< DPTX_APCALL_DEVICE_BUSY_TIMEOUT)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 23)
    def apcall_dev_busy_timeout_ack(self, data):
        self.log(f"> DPTX_APCALL_DEVICE_BUSY_TIMEOUT)")
        chexdump(data, print_fn=self.log)

    @epic_service_notify(0, 24)
    def apcall_dev_not_started(self, data):
        self.log(f"< APCALL_DEVICE_NOT_STARTED)")
        chexdump(data, print_fn=self.log)
    @epic_service_notify_ack(0, 24)
    def apcall_dev_not_started_ack(self, data):
        self.log(f"> APCALL_DEVICE_NOT_STARTED)")
        chexdump(data, print_fn=self.log)

class DPTXPortService(EPICEp):
    NAME = "dptxport"
    SERVICES = [
        DCPDPTXPortEpicTracer
    ]

class DCPTracer(ASCTracer):
    ENDPOINTS = {
        0x20: SystemService,
        0x21: TestService,
        0x22: DCPExpertService,
        # Disp0 / DCP iboot as used by m1n1 is incompatible with the generic
        # EPICEp tracer, disable it for now
        #0x23: Disp0Service,
        0x24: DPAVController,
        0x25: EPICEp, # dcpav-power-ep
        0x26: DPSACService,
        0x27: DPDevService,
        0x28: DPAVService,
        0x29: AVService,
        0x2a: DPTXPortService, # dcpdptx-port-ep
        0x2b: HDCPService,
        0x2c: EPICEp, # cb-ap-to-dcp-service-ep
        0x2d: RemoteAllocService,
        0x37: DCPEp, # iomfb-link
    }

    def handle_msg(self, direction, r0, r1):
        super().handle_msg(direction, r0, r1)
        #iomon.poll()

# get get the SID used by DCP from the only child of "dart-dcp*"
dcp_sid = u.adt[get_dcp_device(DCPDevType.DART_DCP, dcp_name)][0].reg

dart_dcp_tracer = DARTTracer(hv, get_dcp_device(DCPDevType.DART_DCP, dcp_name))
dart_dcp_tracer.start()

dart_disp0_tracer = DARTTracer(hv, get_dcp_device(DCPDevType.DART_DISP, dcp_name))
dart_disp0_tracer.start()

def readmem_iova(addr, size, readfn):
    try:
        return dart_dcp_tracer.dart.ioread(dcp_sid, addr, size)
    except Exception as e:
        print(e)
        return None

iomon.readmem = readmem_iova

dcp_tracer = DCPTracer(hv, get_dcp_device(DCPDevType.DCP, dcp_name), verbose=1)
dcp_tracer.start(dart_dcp_tracer.dart, stream=dcp_sid)

#dcp_tracer.ep.dcpep.state.dumpfile = open(dcp_name + ".log", "a")



# SPDX-License-Identifier: MIT
from .base import *
from ...utils import *

class IOReportingMessage(Register64):
    TYPE = 63, 52

class IOReporting_GetBuf(IOReportingMessage):
    TYPE = 63, 52, Constant(1)
    SIZE = 51, 44
    DVA = 43, 0

class IOReporting_Start(IOReportingMessage):
    TYPE = 63, 52, Constant(0xc)

class IOReporting_Report(IOReportingMessage):
    TYPE = 63, 52, Constant(0x8)

class ASCIOReportingEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = IOReportingMessage
    SHORT = "iorep"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.iobuffer = None
        self.iobuffer_dva = None

    @msg_handler(1, IOReporting_GetBuf)
    def GetBuf(self, msg):
        if self.iobuffer:
            self.log("WARNING: trying to reset iobuffer!")


        if msg.DVA != 0:
            self.bufsize = 0x1000 * msg.SIZE
            self.iobuffer = self.iobuffer_dva = msg.DVA
            self.log(f"buf prealloc {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
        else:
            self.bufsize = align(0x1000 * msg.SIZE, 0x4000)
            self.iobuffer, self.iobuffer_dva = self.asc.ioalloc(self.bufsize)
            self.log(f"buf {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
            self.send(IOReporting_GetBuf(DVA=self.iobuffer_dva, SIZE=self.bufsize // 0x1000))

        return True

    @msg_handler(0xc, IOReporting_Start)
    def Start(self, msg):
        self.log("start")
        return True

    @msg_handler(8, IOReporting_Report)
    def Init(self, msg):
        self.log("report!")
        buf = self.asc.iface.readmem(self.iobuffer, self.bufsize)
        #chexdump(buf)
        self.send(IOReporting_Report())
        return True



# SPDX-License-Identifier: MIT

import struct
from enum import IntEnum
from ..hv import TraceMode
from ..utils import *
from . import ADTDevTracer
from ..hw.asc import *

class DIR(IntEnum):
    RX = 0
    TX = 1

def msg(message, direction=None, regtype=None, name=None):
    def f(x):
        x.is_message = True
        x.direction = direction
        x.message = message
        x.regtype = regtype
        x.name = name
        return x
    return f

def msg_log(*args, **kwargs):
    def x(self, msg):
        return False
    return msg(*args, **kwargs)(x)

def msg_ign(*args, **kwargs):
    def x(self, msg):
        return True
    return msg(*args, **kwargs)(x)

class EPState(object):
    pass

class EP(object):
    NAME = None
    BASE_MESSAGE = None

    def __init__(self, tracer, epid):
        self.tracer = tracer
        self.epid = epid
        self.present = False
        self.started = False
        self.name = self.NAME or type(self).__name__.lower()
        self.state = EPState()
        self.hv = self.tracer.hv
        self.msgmap = {}
        for name in dir(self):
            i = getattr(self, name)
            if not callable(i) or not getattr(i, "is_message", False):
                continue
            self.msgmap[i.direction, i.message] = getattr(self, name), name, i.regtype

    def log(self, msg):
        self.tracer.log(f"[{self.name}] {msg}")

    def start(self):
        pass

    def handle_msg(self, direction, r0, r1):
        msgtype = None
        if self.BASE_MESSAGE:
            r0 = self.BASE_MESSAGE(r0.value)
            msgtype = r0.TYPE

        handler = None
        name = "<unknown>"
        regtype = None

        msgids = [
            (direction, msgtype),
            (None, msgtype),
            (direction, None),
            (None, None),
        ]

        for msgid in msgids:
            handler, name, regtype = self.msgmap.get(msgid, (None, None, None))
            if handler:
                break

        if regtype is not None:
            r0 = regtype(r0.value)

        if handler:
            if handler.name is not None:
                name = handler.name
            if handler(r0):
                return True

        d = ">" if direction == DIR.TX else "<"
        self.log(f"{d}{msgtype:#x}({name}) {r0.value:016x} ({r0.str_fields()})")
        return True

class EPContainer(object):
    pass

class BaseASCTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.SYNC

    REGMAPS = [ASCRegs, None]
    NAMES = ["asc", None]

    ENDPOINTS = {}

    def w_OUTBOX_CTRL(self, val):
        self.log(f"OUTBOX_CTRL = {val!s}")

    def w_INBOX_CTRL(self, val):
        self.log(f"INBOX_CTRL = {val!s}")

    def w_CPU_CONTROL(self, val):
        self.log(f"CPU_CONTROL = {val!s}")

    def w_INBOX1(self, inbox1):
        inbox0 = self.asc.cached.INBOX0.reg
        if self.verbose >= 2:
            self.log(f"SEND: {inbox0.value:016x}:{inbox1.value:016x} " +
                    f"{inbox0.str_fields()} | {inbox1.str_fields()}")
        self.handle_msg(DIR.TX, inbox0, inbox1)

    def r_OUTBOX1(self, outbox1):
        outbox0 = self.asc.cached.OUTBOX0.reg
        if self.verbose >= 2:
            self.log(f"RECV: {outbox0.value:016x}:{outbox1.value:016x} " +
                    f"{outbox0.str_fields()} | {outbox1.str_fields()}")
        self.handle_msg(DIR.RX, outbox0, outbox1)

    def init_state(self):
        self.state.ep = {}

    def handle_msg(self, direction, r0, r1):
        if r1.EP in self.epmap:
            if self.epmap[r1.EP].handle_msg(direction, r0, r1):
                return

        d = ">" if direction == DIR.TX else "<"
        self.log(f"{d}ep:{r1.EP:02x} {r0.value:016x} ({r0.str_fields()})")

    def ioread(self, dva, size):
        if self.dart:
            return self.dart.ioread(self.stream, dva & 0xFFFFFFFFF, size)
        else:
            return self.hv.iface.readmem(dva, size)

    def iowrite(self, dva, data):
        if self.dart:
            return self.dart.iowrite(self.stream, dva & 0xFFFFFFFFF, data)
        else:
            return self.hv.iface.writemem(dva, data)

    def start(self, dart=None, stream=0):
        super().start()
        self.dart = dart
        self.stream = stream
        self.msgmap = {}
        for name in dir(self):
            i = getattr(self, name)
            if not callable(i) or not getattr(i, "is_message", False):
                continue
            self.msgmap[i.direction, i.endpoint, i.message] = getattr(self, name), name, i.regtype

        self.epmap = {}
        self.ep = EPContainer()
        for cls in type(self).mro():
            eps = getattr(cls, "ENDPOINTS", None)
            if eps is None:
                break
            for k, v in eps.items():
                if k in self.epmap:
                    continue
                ep = v(self, k)
                ep.dart = dart
                ep.stream = stream
                self.epmap[k] = ep
                if k in self.state.ep:
                    ep.state.__dict__.update(self.state.ep[k])
                self.state.ep[k] = ep.state.__dict__
                if getattr(self.ep, ep.name, None):
                    ep.name = f"{ep.name}{k:02x}"
                setattr(self.ep, ep.name, ep)
                ep.start()

# System endpoints

## Management endpoint

from ..fw.asc.mgmt import ManagementMessage, Mgmt_EPMap, Mgmt_EPMap_Ack, Mgmt_StartEP, Mgmt_SetAPPower, Mgmt_SetIOPPower, Mgmt_IOPPowerAck

class Management(EP):
    BASE_MESSAGE = ManagementMessage

    HELLO =     msg_log(1, DIR.RX)
    HELLO_ACK = msg_log(2, DIR.TX)

    @msg(5, DIR.TX, Mgmt_StartEP)
    def StartEP(self, msg):
        ep = self.tracer.epmap.get(msg.EP, None)
        if ep:
            ep.started = True
            self.log(f"  Starting endpoint #{msg.EP:#02x} ({ep.name})")
        else:
            self.log(f"  Starting endpoint #{msg.EP:#02x}")
        #return True

    Init = msg_log(6, DIR.TX)

    @msg(8, DIR.RX, Mgmt_EPMap)
    def EPMap(self, msg):
        for i in range(32):
            if msg.BITMAP & (1 << i):
                epno = 32 * msg.BASE + i
                ep = self.tracer.epmap.get(epno, None)
                if ep:
                    ep.present = True
                    self.log(f"  Adding endpoint #{epno:#02x} ({ep.name})")
                else:
                    self.log(f"  Adding endpoint #{epno:#02x}")

    EPMap_Ack = msg_log(8, DIR.TX, Mgmt_EPMap_Ack)

    SetIOPPower = msg_log(6, DIR.TX, Mgmt_SetIOPPower)
    SetIOPPowerAck = msg_log(7, DIR.TX, Mgmt_IOPPowerAck)

    SetAPPower = msg_log(0x0b, DIR.TX, Mgmt_SetAPPower)
    SetAPPowerAck = msg_log(0x0b, DIR.RX, Mgmt_SetAPPower)

## Syslog endpoint

from ..fw.asc.syslog import SyslogMessage, Syslog_Init, Syslog_GetBuf, Syslog_Log

class Syslog(EP):
    BASE_MESSAGE = SyslogMessage

    @msg(8, DIR.RX, Syslog_Init)
    def Init(self, msg):
        self.state.count = msg.COUNT
        self.state.entrysize = msg.ENTRYSIZE

    @msg(1, DIR.RX, Syslog_GetBuf)
    def GetBuf(self, msg):
        if msg.DVA:
            self.state.syslog_buf = msg.DVA

    @msg(1, DIR.TX, Syslog_GetBuf)
    def GetBuf_Ack(self, msg):
        self.state.syslog_buf = msg.DVA

    @msg(5, DIR.RX, Syslog_Log)
    def Log(self, msg):
        buf = self.state.syslog_buf
        stride = 0x20 + self.state.entrysize
        log = self.tracer.ioread(buf + msg.INDEX * stride, stride)
        hdr, unk, context, logmsg = struct.unpack(f"<II24s{self.state.entrysize}s", log)
        context = context.split(b"\x00")[0].decode("ascii")
        logmsg = logmsg.split(b"\x00")[0].decode("ascii").rstrip("\n")
        self.log(f"* [{context}]{logmsg}")
        return True

    Log_Ack = msg_ign(5, DIR.TX, Syslog_Log)

class ASCTracer(BaseASCTracer):
    ENDPOINTS = {
        0: Management,
        #1: CrashLog,
        2: Syslog,
        #3: KDebug,
        #4: IOReporting,
    }



# SPDX-License-Identifier: MIT
from construct import *
import cv2
import struct
from termcolor import colored
import time
from .isp_vid import ISPFrame

class ISPTerminalChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)

    def dump(self):
        msgs = self.read_all_msgs()
        for n, msg in enumerate(msgs):
            if (msg.valid()):
                data = self.isp.ioread(msg.arg0 & ~3, msg.arg1) # iova, size
                s = data.decode().strip().replace('\n', '').replace('\r', '')
                if (s):
                    print("(%d): ISPASC: " % (n) + s)


class ISPIOChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)


class ISPDebugChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)


class ISPBufH2TChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)


class ISPBufT2HChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)

    def _handle(self, req):
        assert((req.arg1 > 0x0))
        self.log("RX: REQ: [iova: 0x%x, size: 0x%x, flag: 0x%x]" % (req.arg0, req.arg1, req.arg2))

        if (req.arg2 == 0x10000000):
            self.push_frame(req)
        else:
            self.log(colored("Warning: frame dropped", "red"))

        # RX: REQ: [0x3a54000, 0x280, 0x10000000]  # iova, size, flag
        # TX: RSP: [0x3a54001, 0x000, 0x80000000]  # iova, zero, flag
        rsp = ISPChannelMessage.build(
            arg0 = req.arg0 | 0x1,  # iova
            arg1 = 0x0,  # zero out size
            arg2 = 0x80000000,  # done flag
        )
        self.log("TX: RSP: [iova: 0x%x, ack: 0x%x, flag: 0x%x]" % (rsp.arg0, rsp.arg1, rsp.arg2))
        return rsp

    def push_frame(self, req):
        frame = ISPFrame(self.isp, req)
        self.isp.frames.append(frame)
        self.log(colored("Pushing %s" % (str(frame)), "green"))
        cv2.imshow('frame', frame.process())
        cv2.waitKey(50)


class ISPSharedMallocChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)

    def _handle(self, req):
        rsp = None

        # Two request types: malloc & free
        if (req.arg0 == 0):
            # On malloc request, arg0 (iova) is zeroed
            # alloc the requested size (arg1) & fill out the new iova
            # RX: REQ: [0x0, 0x18000, 0x4c4f47]  # zero, size, name
            # TX: RSP: [0x3a28001, 0x0, 0xc]  # iova, zero, index
            assert((req.arg1 > 0x0))  # needs size
            try:  # sometimes it sends garbage
                name = struct.pack(">q", req.arg2).decode()
            except UnicodeDecodeError as e:
                name = ("%04x" % (req.arg2 & 0xffff)).upper()
            self.log("SM: RX: Malloc REQ: [size: 0x%x, name: %s]" % (req.arg1, name))

            surf = self.isp.mmger.alloc_size(req.arg1, name=name)
            rsp = ISPChannelMessage.build(
                    arg0 = surf.iova | 0x1,  # fill with new iova
                    arg1 = 0x0, # zero out size as ack
                    arg2 = surf.index, # index for (my) reference
            )
            self.log("SM: TX: Malloc RSP: [iova: 0x%x, index: 0x%x]" % (rsp.arg0, rsp.arg2))

        else:   # On free request, arg0 is the iova to free (duh)
            # RX: REQ: [0x81664000, 0x0, 0x0]  # iova, zero, zero
            # TX: RSP: [0x81664001, 0x0, 0x0]  # iova, zero, zero
            assert((req.arg1 == 0x0) and (req.arg2 == 0x0))  # can't have size/index
            self.log("SM: RX: Free REQ: [iova: 0x%x]" % (req.arg0))

            surf = self.isp.mmger.index_iova(req.arg0)
            if not surf: raise ValueError("shit")
            self.isp.mmger.free_surf(surf)

            rsp = ISPChannelMessage.build(
                    arg0 = req.arg0 | 0x1,  # flag freed iova
                    arg1 = 0x0, # zero out size
                    arg2 = 0x0, # zero out index
            )
            self.log("SM: TX: Free RSP: [iova: 0x%x]" % (rsp.arg0))

        return rsp


class ISPIOT2HChannel(ISPChannel):
    def __init__(self, isp, x, **kwargs):
            super().__init__(isp, x.name, x.type, x.src, x.num, x.iova, **kwargs)


class ISPChannelTable:
    def __init__(self, isp, description):
        self.isp = isp
        for desc in description:
            name = desc.name
            if (name == "TERMINAL"): self.terminal = ISPTerminalChannel(isp, desc, abbrev="TM")
            if (name == "IO"): self.io = ISPIOChannel(isp, desc, abbrev="IO")
            if (name == "DEBUG"): self.debug = ISPDebugChannel(isp, desc, abbrev="DG")
            if (name == "BUF_H2T"): self.bufh2t = ISPBufH2TChannel(isp, desc, abbrev="BR")
            if (name == "BUF_T2H"): self.buft2h = ISPBufT2HChannel(isp, desc, abbrev="BT")
            if (name == "SHAREDMALLOC"): self.sharedmalloc = ISPSharedMallocChannel(isp, desc, abbrev="SM")
            if (name == "IO_T2H"): self.iot2h = ISPIOT2HChannel(isp, desc, abbrev="IT")
        self.channels = [self.terminal, self.io, self.debug, self.bufh2t, self.buft2h, self.sharedmalloc, self.iot2h]
        assert(all(chan for chan in self.channels))

    def name2chan(self, name):
        if (name == "TERMINAL"): return self.terminal
        if (name == "IO"): return self.io
        if (name == "DEBUG"): return self.debug
        if (name == "BUF_H2T"): return self.bufh2t
        if (name == "BUF_T2H"): return self.buft2h
        if (name == "SHAREDMALLOC"): return self.sharedmalloc
        if (name == "IO_T2H"): return self.iot2h

    def dump(self):
        for chan in self.channels:
            if (chan.name != "TERMINAL"):
                chan.dump()



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum


class R_STATUS(Register32):
    DONE = 0
    TIMEOUT = 1
    RD_BUF_OVERFLOW = 2
    WR_BUF_OVERFLOW = 3
    CODEC_BUF_OVERFLOW = 4
    SOME_KIND_OF_MACROBLOCK_SIZE_ERROR = 5
    AXI_ERROR = 6
    UNKNOWN_FLAG = 7


class E_CODEC(IntEnum):
    _444 = 0
    _422 = 1
    _411 = 2
    _420 = 3
    _400 = 4


class R_CODEC(Register32):
    CODEC = 2, 0, E_CODEC


class E_ENCODE_PIXEL_FORMAT(IntEnum):
    RGB101010 = 0
    YUV10_linear = 1
    RGB888 = 2
    RGB565 = 3
    YUV_planar = 4      # partially tested, details not understood
    YUV_linear = 5      # partially tested, details not understood


class R_ENCODE_PIXEL_FORMAT(Register32):
    FORMAT = 4, 0, E_ENCODE_PIXEL_FORMAT


class E_SCALE(IntEnum):
    DIV1 = 0
    DIV2 = 1
    DIV4 = 2
    DIV8 = 3


class R_SCALE_FACTOR(Register32):
    SCALE = 1, 0, E_SCALE


class E_DECODE_PIXEL_FORMAT(IntEnum):
    YUV444_planar = 0
    YUV422_planar = 1
    YUV420_planar = 2
    YUV422_linear = 3
    _YUV10_broken_doesnt_work = 4
    RGBA8888 = 5
    RGB565 = 6
    _RGB101010_broken_doesnt_work = 7


class R_DECODE_PIXEL_FORMAT(Register32):
    FORMAT = 3, 0, E_DECODE_PIXEL_FORMAT


class E_JPEG_IO_FLAGS_SUBSAMPLING(IntEnum):
    _444 = 0
    _422 = 1
    _420 = 2
    _400 = 3
    FOUR_COMPONENTS_MODE = 4
    _411_BROKEN = 6


class R_JPEG_IO_FLAGS(Register32):
    SUBSAMPLING_MODE = 2, 0, E_JPEG_IO_FLAGS_SUBSAMPLING
    # not sure what this is supposed to do
    MAKE_DECODE_WORK_BREAK_ENCODE = 3
    OUTPUT_MACROBLOCKS_UNFLIPPED_H = 4
    OUTPUT_8BYTE_CHUNKS_CORRECTLY = 5


class R_JPEG_OUTPUT_FLAGS(Register32):
    # bit0 doesn't seem to do anything
    SKIP_HEADERS = 1            # output only SOS/EOI, no SOI/DQT/SOF0/DHT
    OUTPUT_SOF0_AFTER_DHT = 2   # output SOF0 after DHT instead of before it
    # bit3 doesn't seem to do anything
    COMPRESS_WORSE = 4          # not sure exactly what this does


class R_QTBL_SEL(Register32):
    COMPONENT0 = 1, 0
    COMPONENT1 = 3, 2
    COMPONENT2 = 5, 4
    COMPONENT3 = 7, 6     # guessed


class JPEGRegs(RegMap):
    REG_0x0 = 0x0, Register32
    REG_0x4 = 0x4, Register32
    MODE = 0x8, Register32
    REG_0xc = 0xc, Register32

    REG_0x10 = 0x10, Register32
    REG_0x14 = 0x14, Register32
    REG_0x18 = 0x18, Register32
    # REG_0x1c = 0x1c, Register32

    REG_0x20 = 0x20, Register32
    STATUS = 0x24, R_STATUS

    CODEC = 0x28, R_CODEC

    REG_0x2c = 0x2c, Register32
    REG_0x30 = 0x30, Register32
    REG_0x34 = 0x34, Register32
    # this changes the output drastically if set to 1 for decode
    # breaks encode if not set to 1
    REG_0x38 = 0x38, Register32

    # not sure what the difference is. siting? type2 seems to win over type1
    CHROMA_HALVE_H_TYPE1 = 0x3c, Register32
    CHROMA_HALVE_H_TYPE2 = 0x40, Register32
    CHROMA_HALVE_V_TYPE1 = 0x44, Register32
    CHROMA_HALVE_V_TYPE2 = 0x48, Register32

    # if double and quadruple both set --> double
    CHROMA_DOUBLE_H = 0x4c, Register32
    CHROMA_QUADRUPLE_H = 0x50, Register32
    CHROMA_DOUBLE_V = 0x54, Register32

    # details not fully understood yet
    PX_USE_PLANE1 = 0x58, Register32
    PX_TILES_W = 0x5c, Register32
    PX_TILES_H = 0x60, Register32
    PX_PLANE0_WIDTH = 0x64, Register32
    PX_PLANE0_HEIGHT = 0x68, Register32
    PX_PLANE0_TILING_H = 0x6c, Register32
    PX_PLANE0_TILING_V = 0x70, Register32
    PX_PLANE0_STRIDE = 0x74, Register32
    PX_PLANE1_WIDTH = 0x78, Register32
    PX_PLANE1_HEIGHT = 0x7c, Register32
    PX_PLANE1_TILING_H = 0x80, Register32
    PX_PLANE1_TILING_V = 0x84, Register32
    PX_PLANE1_STRIDE = 0x88, Register32

    INPUT_START1 = 0x8c, Register32
    INPUT_START2 = 0x90, Register32
    REG_0x94 = 0x94, Register32
    REG_0x98 = 0x98, Register32
    INPUT_END = 0x9c, Register32

    OUTPUT_START1 = 0xa0, Register32
    OUTPUT_START2 = 0xa4, Register32
    OUTPUT_END = 0xa8, Register32

    MATRIX_MULT = irange(0xAC, 11, 4), Register32
    DITHER = irange(0xD8, 10, 4), Register32

    ENCODE_PIXEL_FORMAT = 0x100, R_ENCODE_PIXEL_FORMAT
    # RGB888: R, G, B = byte pos
    # RGB101010: R, G, B = 0/1/2 = low/mid/high bits
    # RGB565: R, G, B = 0/1/2 = low/mid/high bits
    # YUV10: Y, U, V = 0/1/2 = low/mid/high bits
    # YUV linear: Y0 Cb Cr Y1 = byte pos
    # YUV planar: Y U V = 0 for Y, 0/1 for U/V indicating position somehow
    ENCODE_COMPONENT0_POS = 0x104, Register32
    ENCODE_COMPONENT1_POS = 0x108, Register32
    ENCODE_COMPONENT2_POS = 0x10c, Register32
    ENCODE_COMPONENT3_POS = 0x110, Register32

    CONVERT_COLOR_SPACE = 0x114, Register32

    REG_0x118 = 0x118, Register32
    REG_0x11c = 0x11c, Register32

    REG_0x120 = 0x120, Register32

    # details not understood yet
    TILING_ENABLE = 0x124, Register32
    TILING_PLANE0 = 0x128, Register32
    TILING_PLANE1 = 0x12c, Register32

    DECODE_MACROBLOCKS_W = 0x130, Register32
    DECODE_MACROBLOCKS_H = 0x134, Register32
    RIGHT_EDGE_PIXELS = 0x138, Register32
    BOTTOM_EDGE_PIXELS = 0x13c, Register32
    RIGHT_EDGE_SAMPLES = 0x140, Register32
    BOTTOM_EDGE_SAMPLES = 0x144, Register32

    SCALE_FACTOR = 0x148, R_SCALE_FACTOR

    DECODE_PIXEL_FORMAT = 0x14c, R_DECODE_PIXEL_FORMAT
    # 0 = Cb Y'0 Cr Y'1     1 = Y'0 Cb Y'1 Cr
    YUV422_ORDER = 0x150, Register32
    # 0 = BGRA              1 = RGBA
    RGBA_ORDER = 0x154, Register32
    RGBA_ALPHA = 0x158, Register32

    PLANAR_CHROMA_HALVING = 0x15c, Register32

    REG_0x160 = 0x160, Register32
    REG_0x164 = 0x164, Register32
    # REG_0x168 = 0x168, Register32
    REG_0x16c = 0x16c, Register32

    REG_0x170 = 0x170, Register32
    # REG_0x174 = 0x174, Register32
    PERFCOUNTER = 0x178, Register32
    # REG_0x17c = 0x17c, Register32

    # REG_0x180 = 0x180, Register32
    TIMEOUT = 0x184, Register32
    HWREV = 0x188, Register32

    REG_0x18c = 0x18c, Register32
    REG_0x190 = 0x190, Register32
    REG_0x194 = 0x194, Register32
    REG_0x198 = 0x198, Register32
    REG_0x19c = 0x19c, Register32

    ENABLE_RST_LOGGING = 0x1a0, Register32
    RST_LOG_ENTRIES = 0x1a4, Register32

    REG_0x1a8 = 0x1a8, Register32
    REG_0x1ac = 0x1ac, Register32
    REG_0x1b0 = 0x1b0, Register32

    REG_0x1b4 = 0x1b4, Register32
    REG_0x1b8 = 0x1b8, Register32
    REG_0x1bc = 0x1bc, Register32

    REG_0x1c0 = 0x1c0, Register32
    REG_0x1c4 = 0x1c4, Register32

    REG_0x1c8 = 0x1c8, Register32

    REG_0x1cc = 0x1cc, Register32
    REG_0x1d0 = 0x1d0, Register32
    REG_0x1d4 = 0x1d4, Register32
    REG_0x1d8 = 0x1d8, Register32

    REG_0x1dc = 0x1dc, Register32
    REG_0x1e0 = 0x1e0, Register32
    REG_0x1e4 = 0x1e4, Register32
    REG_0x1e8 = 0x1e8, Register32

    REG_0x1ec = 0x1ec, Register32
    REG_0x1f0 = 0x1f0, Register32
    REG_0x1f4 = 0x1f4, Register32
    REG_0x1f8 = 0x1f8, Register32

    REG_0x1fc = 0x1fc, Register32
    REG_0x200 = 0x200, Register32

    REG_0x204 = 0x204, Register32
    REG_0x208 = 0x208, Register32

    REG_0x20c = 0x20c, Register32
    REG_0x210 = 0x210, Register32
    REG_0x214 = 0x214, Register32
    REG_0x218 = 0x218, Register32

    REG_0x21c = 0x21c, Register32
    REG_0x220 = 0x220, Register32

    REG_0x224 = 0x224, Register32
    REG_0x228 = 0x228, Register32

    REG_0x22c = 0x22c, Register32
    REG_0x230 = 0x230, Register32
    REG_0x234 = 0x234, Register32

    REG_0x238 = 0x238, Register32
    REG_0x23c = 0x23c, Register32
    REG_0x240 = 0x240, Register32
    REG_0x244 = 0x244, Register32
    REG_0x248 = 0x248, Register32

    REG_0x24c = 0x24c, Register32
    REG_0x250 = 0x250, Register32
    REG_0x254 = 0x254, Register32
    REG_0x258 = 0x258, Register32
    REG_0x25c = 0x25c, Register32

    REG_0x260 = 0x260, Register32
    REG_0x264 = 0x264, Register32
    REG_0x268 = 0x268, Register32
    REG_0x26c = 0x26c, Register32

    REG_0x280 = 0x280, Register32

    JPEG_IO_FLAGS = 0x1000, R_JPEG_IO_FLAGS
    REG_0x1004 = 0x1004, Register32
    REG_0x1008 = 0x1008, Register32
    QTBL_SEL = 0x100c, R_QTBL_SEL

    # fixme what _exactly_ does this control
    HUFFMAN_TABLE = 0x1010, Register32
    RST_INTERVAL = 0x1014, Register32     # 16 bits effective
    JPEG_HEIGHT = 0x1018, Register32
    JPEG_WIDTH = 0x101c, Register32

    COMPRESSED_BYTES = 0x1020, Register32
    JPEG_OUTPUT_FLAGS = 0x1024, R_JPEG_OUTPUT_FLAGS
    REG_0x1028 = 0x1028, Register32
    REG_0x102c = 0x102c, Register32

    BITSTREAM_CORRUPTION = 0x1030, Register32
    # REG_0x1034 = 0x1034, Register32
    # REG_0x1038 = 0x1038, Register32
    # REG_0x103c = 0x103c, Register32

    REG_0x1080 = 0x1080, Register32
    REG_0x1084 = 0x1084, Register32
    # REG_0x1088 = 0x1088, Register32
    REG_0x108c = 0x108c, Register32
    REG_0x1090 = 0x1090, Register32

    SHIKINO_VERSION_MAGIC0 = 0x10e0, Register32
    SHIKINO_VERSION_MAGIC1 = 0x10e4, Register32
    SHIKINO_VERSION_MAGIC2 = 0x10e8, Register32
    SHIKINO_VERSION_MAGIC3 = 0x10ec, Register32
    SHIKINO_VERSION_MAGIC4 = 0x10f0, Register32
    # REG_0x10f4 = 0x10f4, Register32
    # REG_0x10f8 = 0x10f8, Register32
    # REG_0x10fc = 0x10fc, Register32

    QTBL = irange(0x1100, 64, 4), Register32

    # todo what's the format?
    RSTLOG = irange(0x2000, 1024, 4), Register32



# SPDX-License-Identifier: MIT

from construct import *
from construct.core import Int16ul, Int32ul, Int64ul, Int8ul

from m1n1.hv import TraceMode
from m1n1.utils import *
from m1n1.trace import ADTDevTracer
from m1n1.trace.asc import ASCRegs
from m1n1.trace.asc import ASCTracer

ASCTracer = ASCTracer._reloadcls()

class NVMERegs(RegMap):
    APPLE_NVMMU_NUM = 0x28100, Register32
    APPLE_NVMMU_BASE_ASQ = 0x28108, Register32
    APPLE_NVMMU_BASE_ASQ1 = 0x2810C, Register32
    APPLE_NVMMU_BASE_IOSQ = 0x28110, Register32
    APPLE_NVMMU_BASE_IOSQ1 = 0x28114, Register32
    APPLE_NVMMU_TCB_INVAL = 0x28118, Register32
    APPLE_NVMMU_TCB_STAT = 0x28120, Register32
    APPLE_ANS2_LINEAR_SQ_CTRL = 0x24908, Register32
    APPLE_ANS2_UNKNOWN_CTRL = 0x24008, Register32
    APPLE_ANS2_BOOT_STATUS = 0x1300, Register32
    APPLE_ANS2_MAX_PEND_CMDS_CTRL = 0x1210, Register32
    APPLE_ANS2_LINEAR_ASQ_DB = 0x2490C, Register32
    APPLE_ANS2_LINEAR_IOSQ_DB = 0x24910, Register32

    NVME_REG_CAP = 0x0000, Register32
    NVME_REG_VS = 0x0008, Register32
    NVME_REG_INTMS = 0x000C, Register32
    NVME_REG_INTMC = 0x0010, Register32
    NVME_REG_CC = 0x0014, Register32
    NVME_REG_CSTS = 0x001C, Register32
    NVME_REG_NSSR = 0x0020, Register32
    NVME_REG_AQA = 0x0024, Register32
    NVME_REG_ASQ = 0x0028, Register32
    NVME_REG_ASQ1 = 0x002C, Register32
    NVME_REG_ACQ = 0x0030, Register32
    NVME_REG_CMBLOC = 0x0038, Register32
    NVME_REG_CMBSZ = 0x003C, Register32
    NVME_REG_BPINFO = 0x0040, Register32
    NVME_REG_BPRSEL = 0x0044, Register32
    NVME_REG_BPMBL = 0x0048, Register32
    NVME_REG_CMBMSC = 0x0050, Register32
    NVME_REG_PMRCAP = 0x0E00, Register32
    NVME_REG_PMRCTL = 0x0E04, Register32
    NVME_REG_PMRSTS = 0x0E08, Register32
    NVME_REG_PMREBS = 0x0E0C, Register32
    NVME_REG_PMRSWTP = 0x0E10, Register32
    NVME_REG_DBS = 0x1000, Register32
    NVME_REG_DBS_ASQ = 0x1004, Register32
    NVME_REG_DBS_IOSQ = 0x100C, Register32


AppleTunnelSetTime = Struct(
    "unk" / Int32ul,
    "unix_timestamp" / Int32ul,
    "time_0" / Int64ul,
    "time_1" / Int64ul,
)

NVMECommand = Struct(
    "opcode" / Int8ul,
    "flags" / Int8ul,
    "command_id" / Int16ul,
    "nsid" / Int32ul,
    "cdw0" / Int32ul,
    "cdw1" / Int32ul,
    "metadata" / Int64ul,
    "prp1" / Int64ul,
    "prp2" / Int64ul,
    "cdw10" / Int32ul,
    "cdw11" / Int32ul,
    "cdw12" / Int32ul,
    "cdw13" / Int32ul,
    "cdw14" / Int32ul,
    "cdw16" / Int32ul,
)

NVME_IO_COMMANDS = {
    0x00: "nvme_cmd_flush",
    0x01: "nvme_cmd_write",
    0x02: "nvme_cmd_read",
    0x04: "nvme_cmd_write_uncor",
    0x05: "nvme_cmd_compare",
    0x08: "nvme_cmd_write_zeroes",
    0x09: "nvme_cmd_dsm",
    0x0C: "nvme_cmd_verify",
    0x0D: "nvme_cmd_resv_register",
    0x0E: "nvme_cmd_resv_report",
    0x11: "nvme_cmd_resv_acquire",
    0x15: "nvme_cmd_resv_release",
    0x79: "nvme_cmd_zone_mgmt_send",
    0x7A: "nvme_cmd_zone_mgmt_recv",
    0x7D: "nvme_cmd_zone_append",
}

NVME_ADMIN_COMMANDS = {
    0x00: "nvme_admin_delete_sq",
    0x01: "nvme_admin_create_sq",
    0x02: "nvme_admin_get_log_page",
    0x04: "nvme_admin_delete_cq",
    0x05: "nvme_admin_create_cq",
    0x06: "nvme_admin_identify",
    0x08: "nvme_admin_abort_cmd",
    0x09: "nvme_admin_set_features",
    0x0A: "nvme_admin_get_features",
    0x0C: "nvme_admin_async_event",
    0x0D: "nvme_admin_ns_mgmt",
    0x10: "nvme_admin_activate_fw",
    0x11: "nvme_admin_download_fw",
    0x14: "nvme_admin_dev_self_test",
    0x15: "nvme_admin_ns_attach",
    0x18: "nvme_admin_keep_alive",
    0x19: "nvme_admin_directive_send",
    0x1A: "nvme_admin_directive_recv",
    0x1C: "nvme_admin_virtual_mgmt",
    0x1D: "nvme_admin_nvme_mi_send",
    0x1E: "nvme_admin_nvme_mi_recv",
    0x7C: "nvme_admin_dbbuf",
    0x80: "nvme_admin_format_nvm",
    0x81: "nvme_admin_security_send",
    0x82: "nvme_admin_security_recv",
    0x84: "nvme_admin_sanitize_nvm",
    0x86: "nvme_admin_get_lba_status",
    0xC0: "nvme_admin_vendor_start",
}

APPLE_TUNNEL_CMDS = {0x06: "set_time", 0x38: "get_nand_id", 0xBA: "get_nand_geometry"}

NVMMUTcb = Struct(
    "opcode" / Int8ul,
    "dma_flags" / Int8ul,
    "command_id" / Int8ul,
    "unk0" / Int8ul,
    "length" / Int32ul,
    "unk1a" / Int64ul,
    "unk1b" / Int64ul,
    "prp0" / Int64ul,
    "prp1" / Int64ul,
    "unk2a" / Int64ul,
    "unk2b" / Int64ul,
    # aes_iv, u8[8]
    # aes_data, u8[64]
)


class NVMETracer(ASCTracer):
    DEFAULT_MODE = TraceMode.SYNC

    REGMAPS = [ASCRegs, None, None, NVMERegs]
    NAMES = ["asc", None, None, "nvme"]

    ENDPOINTS = {}

    def init_state(self):
        self.state.ep = {}
        self.state.cmd_cache = {}
        self.state.nvmmu_asq_base = None
        self.state.nvmmu_iosq_base = None
        self.state.asq = None

    def r_APPLE_NVMMU_TCB_STAT(self, r):
        pass

    def w_APPLE_NVMMU_BASE_ASQ(self, r):
        self.state.nvmmu_asq_base = r.value

    def w_APPLE_NVMMU_BASE_ASQ1(self, r):
        self.state.nvmmu_asq_base |= r.value << 32

    def w_APPLE_NVMMU_BASE_IOSQ(self, r):
        self.state.nvmmu_iosq_base = r.value

    def w_APPLE_NVMMU_BASE_IOSQ1(self, r):
        self.state.nvmmu_iosq_base |= r.value << 32

    def w_NVME_REG_ASQ(self, r):
        self.state.asq = r.value

    def w_NVME_REG_ASQ1(self, r):
        self.state.asq |= r.value << 32

    def w_APPLE_ANS2_LINEAR_ASQ_DB(self, r):
        tag = r.value
        cmd = NVMECommand.parse(self.hv.iface.readmem(self.state.asq + 64 * tag, 0x40))
        tcb = NVMMUTcb.parse(
            self.hv.iface.readmem(self.state.nvmmu_asq_base + 0x80 * tag, 0x80)
        )

        self.state.cmd_cache[tag] = (True, cmd, tcb)

        if cmd.opcode == 0xD8:
            self.log("apple_tunnel_cmd:")
            self.parse_apple_tunnel_cmd(cmd, False)
            return

        cmdname = NVME_ADMIN_COMMANDS.get(cmd.opcode, "unknown")
        self.log(f"{cmdname}:")
        self.log(f"   {repr(cmd)}")
        self.log(f"   {repr(tcb)}")

        if cmd.opcode == 1:
            self.state.iosq = cmd.prp1

    def w_APPLE_ANS2_LINEAR_IOSQ_DB(self, r):
        tag = r.value
        cmd = NVMECommand.parse(self.hv.iface.readmem(self.state.iosq + 64 * tag, 0x40))
        tcb = NVMMUTcb.parse(
            self.hv.iface.readmem(self.state.nvmmu_iosq_base + 0x80 * tag, 0x80)
        )
        cmdname = NVME_IO_COMMANDS.get(cmd.opcode, "unknown")
        self.log(f"{cmdname}:")
        self.log(f"   {repr(cmd)}")
        self.log(f"   {repr(tcb)}")

        self.state.cmd_cache[tag] = (False, cmd, tcb)

    def parse_apple_tunnel_cmd(self, cmd, done):
        ptr0 = (cmd.cdw12 << 32) | cmd.cdw11
        ptr1 = (cmd.cdw14 << 32) | cmd.cdw13

        data = self.hv.iface.readmem(ptr0, 0x4000)
        if ptr1 > 0:
            data1 = self.hv.iface.readmem(ptr1, 0x4000)

        apple_cmd_opcode = data[12]
        apple_cmd = APPLE_TUNNEL_CMDS.get(apple_cmd_opcode, "Unknown")

        if apple_cmd_opcode == 0x06:
            self.log(
                f"  apple_tunnel_cmd: set_time: {repr(AppleTunnelSetTime.parse(data[0x18:0x30]))}"
            )
        elif apple_cmd_opcode == 0x38:
            self.log(f"  apple_tunnel_cmd: get_nand_id")
            if done:
                self.log(f"    manufacturer id: {hexdump(data1[:8])}")
        else:
            self.log(f"  apple_tunnel_cmd: {apple_cmd} ({apple_cmd_opcode})")
            chexdump(data, print_fn=self.log)
            if ptr1 > 0:
                chexdump(self.hv.iface.readmem(ptr1, 0x4000), print_fn=self.log)

    def w_APPLE_NVMMU_TCB_INVAL(self, r):
        self.log(f"   NVMMU inval for {r.value}")
        tag = r.value
        if tag not in self.state.cmd_cache:
            self.log("     NVMMU tag not found in cmd_cache")
            return

        is_admin, cmd, tcb = self.state.cmd_cache[tag]
        del self.state.cmd_cache[tag]

        if is_admin:
            if cmd.opcode == 0xD8:
                self.log(f"  done apple_tunnel_cmd")
                self.parse_apple_tunnel_cmd(cmd, True)
            else:
                cmdname = NVME_ADMIN_COMMANDS.get(cmd.opcode, "unknown")
                self.log(f"  done {cmdname}")
        else:
            cmdname = NVME_IO_COMMANDS.get(cmd.opcode, "unknown")
            self.log(f"  done {cmdname}")

    def start(self):
        self.state.cmd_cache = {}
        super().start()


NVMETracer = NVMETracer._reloadcls()
nvme_tracer = NVMETracer(hv, "/arm-io/ans", verbose=1)
nvme_tracer.start()

trace_device("/arm-io/sart-ans")
